"""
  Motivated by geometric insights on the loss surface of deep neural networks,
  Fast Geometirc Ensembling (FGE) is an efficient ensemble that uses a
  customized learning rate scheduler to generate base estimators, similar to
  snapshot ensemble.

  Reference:
      T. Garipov, P. Izmailov, D. Podoprikhin et al., Loss Surfaces, Mode
      Connectivity, and Fast Ensembling of DNNs, NeurIPS, 2018.
"""


import torch
import logging
import warnings
import torch.nn as nn
import torch.nn.functional as F

from ._base import BaseModule, BaseClassifier, BaseRegressor
from ._base import torchensemble_model_doc
from .utils import io
from .utils import set_module
from .utils import operator as op
from .utils.logging import get_tb_logger


__all__ = ["FastGeometricClassifier", "FastGeometricRegressor"]


__fit_doc = """
    Parameters
    ----------
    train_loader : torch.utils.data.DataLoader
        A :mod:`DataLoader` container that contains the training data.
    cycle : int, default=4
        The number of cycles used to build each base estimator in the ensemble.
    lr_1 : float, default=5e-2
        ``alpha_1`` in original paper used to adjust the learning rate, also
        serves as the initial learning rate of the internal optimizer.
    lr_2 : float, default=1e-4
        ``alpha_2`` in original paper used to adjust the learning rate, also
        serves as the smallest learning rate of the internal optimizer.
    epochs : int, default=100
        The number of training epochs used to fit the dummy base estimator.
    log_interval : int, default=100
        The number of batches to wait before logging the training status.
    test_loader : torch.utils.data.DataLoader, default=None
        A :mod:`DataLoader` container that contains the evaluating data.

        - If ``None``, no validation is conducted after each real base
          estimator being generated.
        - If not ``None``, the ensemble will be evaluated on this
          dataloader after each base estimator being generated.
    save_model : bool, default=True
        Specify whether to save the model parameters.

        - If test_loader is ``None``, the ensemble fully trained will be
          saved.
        - If test_loader is not ``None``, the ensemble with the best
          validation performance will be saved.
    save_dir : string, default=None
        Specify where to save the model parameters.

        - If ``None``, the model will be saved in the current directory.
        - If not ``None``, the model will be saved in the specified
          directory: ``save_dir``.
"""


def _fast_geometric_model_doc(header, item="fit"):
    """
    Decorator on obtaining documentation for different fast geometric models.
    """

    def get_doc(item):
        """Return selected item"""
        __doc = {"fit": __fit_doc}
        return __doc[item]

    def adddoc(cls):
        doc = [header + "\n\n"]
        doc.extend(get_doc(item))
        cls.__doc__ = "".join(doc)
        return cls

    return adddoc


class _BaseFastGeometric(BaseModule):
    def __init__(
        self, estimator, n_estimators, estimator_args=None, cuda=True
    ):
        super(BaseModule, self).__init__()

        self.base_estimator_ = estimator
        self.n_estimators = n_estimators
        self.estimator_args = estimator_args

        if estimator_args and not isinstance(estimator, type):
            msg = (
                "The input `estimator_args` will have no effect since"
                " `estimator` is already an object after instantiation."
            )
            warnings.warn(msg, RuntimeWarning)

        self.device = torch.device("cuda" if cuda else "cpu")
        self.logger = logging.getLogger()
        self.tb_logger = get_tb_logger()

        self.estimators_ = nn.ModuleList()
        self.use_scheduler_ = False

    def _forward(self, *x):
        """
        Implementation on the internal data forwarding in fast geometric
        ensemble.
        """
        # Average
        results = [estimator(*x) for estimator in self.estimators_]
        output = op.average(results)

        return output

    def _adjust_lr(
        self, optimizer, epoch, i, n_iters, cycle, alpha_1, alpha_2
    ):
        """
        Set the internal learning rate scheduler for fast geometric ensemble.
        Please refer to the original paper for details.
        """

        def scheduler(i):
            t = ((epoch % cycle) + i) / cycle
            if t < 0.5:
                return alpha_1 * (1.0 - 2.0 * t) + alpha_2 * 2.0 * t
            else:
                return alpha_1 * (2.0 * t - 1.0) + alpha_2 * (2.0 - 2.0 * t)

        lr = scheduler(i / n_iters)
        for param_group in optimizer.param_groups:
            param_group["lr"] = lr

        return lr

    @torchensemble_model_doc(
        """Set the attributes on optimizer for Fast Geometric Ensemble.""",
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        self.optimizer_name = optimizer_name
        self.optimizer_args = kwargs

    @torchensemble_model_doc(
        """Set the attributes on scheduler for Fast Geometric Ensemble.""",
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        msg = (
            "The learning rate scheduler for fast geometirc ensemble will"
            " only be used in the first stage on building the dummy base"
            " estimator."
        )
        warnings.warn(msg, UserWarning)

        self.scheduler_name = scheduler_name
        self.scheduler_args = kwargs
        self.use_scheduler_ = True


@torchensemble_model_doc(
    """Implementation on the FastGeometricClassifier.""", "seq_model"
)
class FastGeometricClassifier(_BaseFastGeometric, BaseClassifier):
    @torchensemble_model_doc(
        """Implementation on the data forwarding in FastGeometricClassifier.""",  # noqa: E501
        "classifier_forward",
    )
    def forward(self, *x):
        proba = self._forward(*x)

        return F.softmax(proba, dim=1)

    @torchensemble_model_doc(
        (
            """Set the attributes on optimizer for FastGeometricClassifier. """
            + """Notice that keyword arguments specified here will also be """
            + """used in the ensembling stage except the learning rate.."""
        ),
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name=optimizer_name, **kwargs)

    @torchensemble_model_doc(
        (
            """Set the attributes on scheduler for FastGeometricClassifier. """
            + """Notice that this scheduler will only be used in the stage on """  # noqa: E501
            + """fitting the dummy base estimator."""
        ),
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        super().set_scheduler(scheduler_name=scheduler_name, **kwargs)

    @torchensemble_model_doc(
        """Set the training criterion for FastGeometricClassifier.""",
        "set_criterion",
    )
    def set_criterion(self, criterion):
        super().set_criterion(criterion)

    @_fast_geometric_model_doc(
        """Implementation on the training stage of FastGeometricClassifier.""",  # noqa: E501
        "fit",
    )
    def fit(
        self,
        train_loader,
        cycle=4,
        lr_1=5e-2,
        lr_2=1e-4,
        epochs=100,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):
        self._validate_parameters(epochs, log_interval)
        self.n_outputs = self._decide_n_outputs(train_loader)

        # ====================================================================
        #                Train the dummy estimator (estimator_)
        # ====================================================================

        estimator_ = self._make_estimator()

        # Set the optimizer and scheduler
        optimizer = set_module.set_optimizer(
            estimator_, self.optimizer_name, **self.optimizer_args
        )

        if self.use_scheduler_:
            scheduler = set_module.set_scheduler(
                optimizer, self.scheduler_name, **self.scheduler_args
            )

        # Check the training criterion
        if not hasattr(self, "_criterion"):
            self._criterion = nn.CrossEntropyLoss()

        # Utils
        total_iters = 0

        for epoch in range(epochs):

            # Training
            estimator_.train()
            for batch_idx, elem in enumerate(train_loader):

                data, target = io.split_data_target(elem, self.device)
                batch_size = data[0].size(0)

                optimizer.zero_grad()
                output = estimator_(*data)
                loss = self._criterion(output, target)
                loss.backward()
                optimizer.step()

                # Print training status
                if batch_idx % log_interval == 0:
                    with torch.no_grad():
                        _, predicted = torch.max(output.data, 1)
                        correct = (predicted == target).sum().item()

                        msg = (
                            "Epoch: {:03d} | Batch: {:03d} | Loss: {:.5f} |"
                            " Correct: {:d}/{:d}"
                        )
                        self.logger.info(
                            msg.format(
                                epoch,
                                batch_idx,
                                loss,
                                correct,
                                batch_size,
                            )
                        )
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "fast_geometric/Base_Est/Train_Loss",
                                loss,
                                total_iters,
                            )
                total_iters += 1

            if self.use_scheduler_:
                scheduler.step()

        # ====================================================================
        #                        Generate the ensemble
        # ====================================================================

        # Set the internal optimizer
        estimator_.zero_grad()
        optimizer = set_module.set_optimizer(
            estimator_, self.optimizer_name, **self.optimizer_args
        )

        # Utils
        best_acc = 0.0
        n_iters = len(train_loader)
        updated = False
        epoch = 0

        while len(self.estimators_) < self.n_estimators:

            # Training
            estimator_.train()
            for batch_idx, elem in enumerate(train_loader):

                # Update learning rate
                self._adjust_lr(
                    optimizer, epoch, batch_idx, n_iters, cycle, lr_1, lr_2
                )

                data, target = io.split_data_target(elem, self.device)
                batch_size = data[0].size(0)

                optimizer.zero_grad()
                output = estimator_(*data)
                loss = self._criterion(output, target)
                loss.backward()
                optimizer.step()

                # Print training status
                if batch_idx % log_interval == 0:
                    with torch.no_grad():
                        _, predicted = torch.max(output.data, 1)
                        correct = (predicted == target).sum().item()

                        msg = (
                            "lr: {:.5f} | Epoch: {:03d} | Batch: {:03d} |"
                            " Loss: {:.5f} | Correct: {:d}/{:d}"
                        )
                        self.logger.info(
                            msg.format(
                                optimizer.param_groups[0]["lr"],
                                epoch,
                                batch_idx,
                                loss,
                                correct,
                                batch_size,
                            )
                        )
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "fast_geometric/Ensemble-Est_{}".format(
                                    len(self.estimators_)
                                )
                                + "/Train_Loss",
                                loss,
                                total_iters,
                            )
                total_iters += 1

            # Update the ensemble
            if (epoch % cycle + 1) == cycle // 2:
                base_estimator = self._make_estimator()
                base_estimator.load_state_dict(estimator_.state_dict())
                self.estimators_.append(base_estimator)
                updated = True
                total_iters = 0

                msg = "Save the base estimator with index: {}"
                self.logger.info(msg.format(len(self.estimators_) - 1))

            # Validation after each base estimator being added
            if test_loader and updated:
                self.eval()
                with torch.no_grad():
                    correct = 0
                    total = 0
                    for _, elem in enumerate(test_loader):
                        data, target = io.split_data_target(elem, self.device)
                        output = self.forward(*data)
                        _, predicted = torch.max(output.data, 1)
                        correct += (predicted == target).sum().item()
                        total += target.size(0)
                    acc = 100 * correct / total

                    if acc > best_acc:
                        best_acc = acc
                        if save_model:
                            io.save(self, save_dir, self.logger)

                    msg = (
                        "n_estimators: {} | Validation Acc: {:.3f} %"
                        " | Historical Best: {:.3f} %"
                    )
                    self.logger.info(
                        msg.format(len(self.estimators_), acc, best_acc)
                    )
                    if self.tb_logger:
                        self.tb_logger.add_scalar(
                            "fast_geometric/Ensemble_Est/Validation_Acc",
                            acc,
                            len(self.estimators_),
                        )
                updated = False  # reset the updating flag
            epoch += 1

        if save_model and not test_loader:
            io.save(self, save_dir, self.logger)

    @torchensemble_model_doc(item="classifier_evaluate")
    def evaluate(self, test_loader, return_loss=False):
        return super().evaluate(test_loader, return_loss)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)


@torchensemble_model_doc(
    """Implementation on the FastGeometricRegressor.""", "seq_model"
)
class FastGeometricRegressor(_BaseFastGeometric, BaseRegressor):
    @torchensemble_model_doc(
        """Implementation on the data forwarding in FastGeometricRegressor.""",  # noqa: E501
        "regressor_forward",
    )
    def forward(self, *x):
        pred = self._forward(*x)
        return pred

    @torchensemble_model_doc(
        (
            """Set the attributes on optimizer for FastGeometricRegressor. """
            + """Notice that keyword arguments specified here will also be """
            + """used in the ensembling stage except the learning rate."""
        ),
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name=optimizer_name, **kwargs)

    @torchensemble_model_doc(
        (
            """Set the attributes on scheduler for FastGeometricRegressor. """
            + """Notice that this scheduler will only be used in the stage on """  # noqa: E501
            + """fitting the dummy base estimator."""
        ),
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        super().set_scheduler(scheduler_name=scheduler_name, **kwargs)

    @torchensemble_model_doc(
        """Set the training criterion for FastGeometricRegressor.""",
        "set_criterion",
    )
    def set_criterion(self, criterion):
        super().set_criterion(criterion)

    @_fast_geometric_model_doc(
        """Implementation on the training stage of FastGeometricRegressor.""",  # noqa: E501
        "fit",
    )
    def fit(
        self,
        train_loader,
        cycle=4,
        lr_1=5e-2,
        lr_2=1e-4,
        epochs=100,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):
        self._validate_parameters(epochs, log_interval)
        self.n_outputs = self._decide_n_outputs(train_loader)

        # ====================================================================
        #                Train the dummy estimator (estimator_)
        # ====================================================================

        estimator_ = self._make_estimator()

        # Set the optimizer and scheduler
        optimizer = set_module.set_optimizer(
            estimator_, self.optimizer_name, **self.optimizer_args
        )

        if self.use_scheduler_:
            scheduler = set_module.set_scheduler(
                optimizer, self.scheduler_name, **self.scheduler_args
            )

        # Check the training criterion
        if not hasattr(self, "_criterion"):
            self._criterion = nn.MSELoss()

        # Utils
        total_iters = 0

        for epoch in range(epochs):

            # Training
            estimator_.train()
            for batch_idx, elem in enumerate(train_loader):

                data, target = io.split_data_target(elem, self.device)

                optimizer.zero_grad()
                output = estimator_(*data)
                loss = self._criterion(output, target)
                loss.backward()
                optimizer.step()

                # Print training status
                if batch_idx % log_interval == 0:
                    with torch.no_grad():
                        msg = "Epoch: {:03d} | Batch: {:03d} | Loss: {:.5f}"
                        self.logger.info(msg.format(epoch, batch_idx, loss))
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "fast_geometric/Base_Est/Train_Loss",
                                loss,
                                total_iters,
                            )
                total_iters += 1

            if self.use_scheduler_:
                scheduler.step()

        # ====================================================================
        #                        Generate the ensemble
        # ====================================================================

        # Set the internal optimizer
        estimator_.zero_grad()
        optimizer = set_module.set_optimizer(
            estimator_, self.optimizer_name, **self.optimizer_args
        )

        # Utils
        best_loss = float("inf")
        n_iters = len(train_loader)
        updated = False
        epoch = 0

        while len(self.estimators_) < self.n_estimators:

            # Training
            estimator_.train()
            for batch_idx, elem in enumerate(train_loader):

                # Update learning rate
                self._adjust_lr(
                    optimizer, epoch, batch_idx, n_iters, cycle, lr_1, lr_2
                )

                data, target = io.split_data_target(elem, self.device)

                optimizer.zero_grad()
                output = estimator_(*data)
                loss = self._criterion(output, target)
                loss.backward()
                optimizer.step()

                # Print training status
                if batch_idx % log_interval == 0:
                    with torch.no_grad():
                        msg = "Epoch: {:03d} | Batch: {:03d} | Loss: {:.5f}"
                        self.logger.info(msg.format(epoch, batch_idx, loss))
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "fast_geometric/Ensemble-Est_{}".format(
                                    len(self.estimators_)
                                )
                                + "/Train_Loss",
                                loss,
                                total_iters,
                            )
                total_iters += 1

            # Update the ensemble
            if (epoch % cycle + 1) == cycle // 2:
                base_estimator = self._make_estimator()
                base_estimator.load_state_dict(estimator_.state_dict())
                self.estimators_.append(base_estimator)
                updated = True
                total_iters = 0

                msg = "Save the base estimator with index: {}"
                self.logger.info(msg.format(len(self.estimators_) - 1))

            # Validation after each base estimator being added
            if test_loader and updated:
                self.eval()
                with torch.no_grad():
                    val_loss = 0.0
                    for _, elem in enumerate(test_loader):
                        data, target = io.split_data_target(elem, self.device)
                        output = self.forward(*data)
                        val_loss += self._criterion(output, target)
                    val_loss /= len(test_loader)

                    if val_loss < best_loss:
                        best_loss = val_loss
                        if save_model:
                            io.save(self, save_dir, self.logger)

                    msg = (
                        "Epoch: {:03d} | Validation Loss: {:.5f} |"
                        " Historical Best: {:.5f}"
                    )
                    self.logger.info(msg.format(epoch, val_loss, best_loss))
                    if self.tb_logger:
                        self.tb_logger.add_scalar(
                            "fast_geometric/Ensemble_Est/Validation_Loss",
                            val_loss,
                            len(self.estimators_),
                        )
                updated = False  # reset the updating flag
            epoch += 1

        if save_model and not test_loader:
            io.save(self, save_dir, self.logger)

    @torchensemble_model_doc(item="regressor_evaluate")
    def evaluate(self, test_loader):
        return super().evaluate(test_loader)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)



from itertools import product

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from rich.console import Console

from rrt_ml.environments.car_navigation_bullet_env import *
from rrt_ml.utilities.analytic import *
from rrt_ml.utilities.configs import *


cfg = MasterConfig()
cfg.env.general.seed = 0
cfg.env.general.gui = True
cfg.env.general.stats = True
cfg.env.general.max_timestep = 1000
cfg.env.target.pose = 1.5
cfg.env.car.pose = [0, 0, 0]
cfg.env.car.f_val = 240

cfg.env.car.pose = [0, 0, 0]
cfg.env.target.pose = [0, 0.79, np.pi]

cfg.rrt.rs.curvature = 2.45

# Initialize env and reset master
env = CarNavigationBulletEnv(cfg)
env.reset_master()
env.reset()

rsmpc = MPCRS(cfg)
action = rsmpc._set_rs_path(env._get_observation())

fig, ax = plt.subplots()
ax.plot(rsmpc.rs_xs, rsmpc.rs_ys, linewidth=2)
for ep in range(50):

    # Control loop
    done, obs, info = False, env.reset(), None
    [env.step([0, 1]) for _ in range(50)]
    xs, ys = [], []
    while not done:
        obs, reward, done, info = env.step([1, 1])
        xs.append(obs['observation'][0])
        ys.append(obs['observation'][1])
    ax.plot(xs, ys, 'r--')
    fig.show()



"""
Density modules for estimating density of items in the replay buffer (e.g., states / achieved goals).
"""

import mrl
import numpy as np
from sklearn.neighbors import KernelDensity
from scipy.special import entr
from mrl.replays.online_her_buffer import OnlineHERBuffer
from mrl.utils.networks import MLP
import torch
import torch.nn.functional as F
import os
from mrl.utils.realnvp import RealNVP


class RawKernelDensity(mrl.Module):
  """
  A KDE-based density model for raw items in the replay buffer (e.g., states/goals).
  """
  def __init__(self, item, optimize_every=10, samples=10000, kernel='gaussian', bandwidth=0.1, normalize=True, 
    log_entropy=False, tag='', buffer_name='replay_buffer'):

    super().__init__('{}_kde{}'.format(item, tag), required_agent_modules=[buffer_name], locals=locals())

    self.step = 0
    self.item = item
    self.kde = KernelDensity(kernel=kernel, bandwidth=bandwidth)
    self.optimize_every = optimize_every
    self.samples = samples
    self.kernel = kernel
    self.bandwidth = bandwidth
    self.normalize = normalize
    self.kde_sample_mean = 0.
    self.kde_sample_std = 1.
    self.fitted_kde = None
    self.ready = False
    self.log_entropy = log_entropy
    self.buffer_name = buffer_name

  def _setup(self):
    assert isinstance(getattr(self, self.buffer_name), OnlineHERBuffer)

  def _optimize(self, force=False):
    buffer = getattr(self, self.buffer_name).buffer.BUFF['buffer_' + self.item]
    self.step +=1

    if force or (self.step % self.optimize_every == 0 and len(buffer)):
      self.ready = True
      sample_idxs = np.random.randint(len(buffer), size=self.samples)
      kde_samples = buffer.get_batch(sample_idxs)
      #og_kde_samples = kde_samples

      if self.normalize:
        self.kde_sample_mean = np.mean(kde_samples, axis=0, keepdims=True)
        self.kde_sample_std  = np.std(kde_samples, axis=0, keepdims=True) + 1e-4
        kde_samples = (kde_samples - self.kde_sample_mean) / self.kde_sample_std

      #if self.item == 'ag' and hasattr(self, 'ag_interest') and self.ag_interest.ready:
      #  ag_weights = self.ag_interest.evaluate_disinterest(og_kde_samples)
      #  self.fitted_kde = self.kde.fit(kde_samples, sample_weight=ag_weights.flatten())
      #else:
      self.fitted_kde = self.kde.fit(kde_samples)

      # Now also log the entropy
      if self.log_entropy and hasattr(self, 'logger') and self.step % 250 == 0:
        # Scoring samples is a bit expensive, so just use 1000 points
        num_samples = 1000
        s = self.fitted_kde.sample(num_samples)
        entropy = -self.fitted_kde.score(s)/num_samples + np.log(self.kde_sample_std).sum()
        self.logger.add_scalar('Explore/{}_entropy'.format(self.module_name), entropy, log_every=500)

  def evaluate_log_density(self, samples):
    assert self.ready, "ENSURE READY BEFORE EVALUATING LOG DENSITY"
    return self.fitted_kde.score_samples( (samples  - self.kde_sample_mean) / self.kde_sample_std )

  def evaluate_elementwise_entropy(self, samples, beta=0.):
    """ Given an array of samples, compute elementwise function of entropy of the form:

        elem_entropy = - (p(samples) + beta)*log(p(samples) + beta)

    Args:
      samples: 1-D array of size N
      beta: float, offset entropy calculation

    Returns:
      elem_entropy: 1-D array of size N, elementwise entropy with beta offset
    """
    assert self.ready, "ENSURE READY BEFORE EVALUATING ELEMENT-WISE ENTROPY"
    log_px = self.fitted_kde.score_samples( (samples  - self.kde_sample_mean) / self.kde_sample_std )
    px = np.exp(log_px)
    elem_entropy = entr(px + beta)
    return elem_entropy

  def save(self, save_folder):
    self._save_props(['kde', 'kde_sample_mean', 'kde_sample_std', 'fitted_kde', 'ready'], save_folder)

  def load(self, save_folder):
    self._load_props(['kde', 'kde_sample_mean', 'kde_sample_std', 'fitted_kde', 'ready'], save_folder)


class RawJointKernelDensity(mrl.Module):
  """
  A KDE-based density model for joint raw items in the replay buffer (e.g., behaviour and achieved goals).

  Args:
    item: a list of items in the replay buffer to build a joint density over
  """
  def __init__(self, items, optimize_every=10, samples=10000, kernel='gaussian', bandwidth=0.1, normalize=True, log_entropy=False, tag=''):
    super().__init__('{}_kde{}'.format("".join(items), tag), required_agent_modules=['replay_buffer'], locals=locals())

    self.step = 0
    self.items = items
    self.kde = KernelDensity(kernel=kernel, bandwidth=bandwidth)
    self.optimize_every = optimize_every
    self.samples = samples
    self.kernel = kernel
    self.bandwidth = bandwidth
    self.normalize = normalize
    self.kde_sample_mean = 0.
    self.kde_sample_std = 1.
    self.fitted_kde = None
    self.ready = False
    self.log_entropy = log_entropy

  def _setup(self):
    assert isinstance(self.replay_buffer, OnlineHERBuffer)

  def _optimize(self, force=False):
    buffers = []
    for item in self.items:
      buffers.append(self.replay_buffer.buffer.BUFF['buffer_' + item])
    
    self.step +=1

    if force or (self.step % self.optimize_every == 0 and len(buffers[0])):
      self.ready = True
      sample_idxs = np.random.randint(len(buffers[0]), size=self.samples)
      kde_samples = []
      for buffer in buffers:
        kde_samples.append(buffer.get_batch(sample_idxs))
      
      # Concatenate the items
      kde_samples = np.concatenate(kde_samples, axis=-1)

      if self.normalize:
        self.kde_sample_mean = np.mean(kde_samples, axis=0, keepdims=True)
        self.kde_sample_std  = np.std(kde_samples, axis=0, keepdims=True) + 1e-4
        kde_samples = (kde_samples - self.kde_sample_mean) / self.kde_sample_std
      
      self.fitted_kde = self.kde.fit(kde_samples)

      # Now also log the entropy
      if self.log_entropy and hasattr(self, 'logger') and self.step % 250 == 0:
        # Scoring samples is a bit expensive, so just use 1000 points
        num_samples = 1000
        s = self.fitted_kde.sample(num_samples)
        entropy = -self.fitted_kde.score(s)/num_samples + np.log(self.kde_sample_std).prod()
        self.logger.add_scalar('Explore/{}_entropy'.format(self.module_name), entropy, log_every=500)

  def evaluate_log_density(self, samples):
    assert self.ready, "ENSURE READY BEFORE EVALUATING LOG DENSITY"
    return self.fitted_kde.score_samples( (samples  - self.kde_sample_mean) / self.kde_sample_std )

  def evaluate_elementwise_entropy(self, samples, beta=0.):
    """ Given an array of samples, compute elementwise function of entropy of the form:

        elem_entropy = - (p(samples) + beta)*log(p(samples) + beta)

    Args:
      samples: 1-D array of size N
      beta: float, offset entropy calculation

    Returns:
      elem_entropy: 1-D array of size N, elementwise entropy with beta offset
    """
    assert self.ready, "ENSURE READY BEFORE EVALUATING ELEMENT-WISE ENTROPY"
    log_px = self.fitted_kde.score_samples( (samples  - self.kde_sample_mean) / self.kde_sample_std )
    px = np.exp(log_px)
    elem_entropy = entr(px + beta)
    return elem_entropy

  def save(self, save_folder):
    self._save_props(['kde', 'kde_sample_mean', 'kde_sample_std', 'fitted_kde', 'ready'], save_folder)

  def load(self, save_folder):
    self._load_props(['kde', 'kde_sample_mean', 'kde_sample_std', 'fitted_kde', 'ready'], save_folder)


class RandomNetworkDensity(mrl.Module):
  """
  A random network based ``density'' model for raw items in the replay buffer (e.g., states/goals). The ``density'' is in proportion
  to the error of the learning network.
  Based on https://arxiv.org/abs/1810.12894.
  """
  def __init__(self, item, optimize_every=1, batch_size=256, layers=(256, 256)):

    super().__init__('{}_rnd'.format(item), required_agent_modules=['replay_buffer'], locals=locals())

    self.step = 0
    self.item = item
    self.layers = layers
    self.optimize_every = optimize_every
    self.batch_size = batch_size
    self.tgt_net, self.prd_net, self.optimizer = None, None, None
    self.lazy_load = None

  def _setup(self):
    assert isinstance(self.replay_buffer, OnlineHERBuffer)

  def _init_from_sample(self, x):
    input_size = x.shape[-1]
    self.tgt_net = MLP(input_size, output_size = self.layers[-1], hidden_sizes = self.layers[:-1])
    self.prd_net = MLP(input_size, output_size = self.layers[-1], hidden_sizes = self.layers[:-1])
    if self.config.get('device'):
      self.tgt_net = self.tgt_net.to(self.config.device)
      self.prd_net = self.prd_net.to(self.config.device)
    self.optimizer = torch.optim.SGD(self.prd_net.parameters(), lr=0.1, weight_decay=1e-5)

  def evaluate_log_density(self, samples):
    """Not actually log density, just prediction error"""
    assert self.tgt_net is not None, "ENSURE READY BEFORE EVALUATING LOG DENSITY"

    samples = self.torch(samples)
    tgt = self.tgt_net(samples)
    prd = self.prd_net(samples)
    return self.numpy(-torch.mean((prd - tgt)**2, dim=-1, keepdim=True))

  @property
  def ready(self):
    return self.tgt_net is not None

  def _optimize(self, force=False):
    buffer = self.replay_buffer.buffer.BUFF['buffer_' + self.item]
    self.step +=1

    if force or (self.step % self.optimize_every == 0 and len(buffer)):
      sample_idxs = np.random.randint(len(buffer), size=self.batch_size)
      samples = buffer.get_batch(sample_idxs)

      # lazy load the networks if not yet loaded
      if self.tgt_net is None:
        self._init_from_sample(samples)
        if self.lazy_load is not None:
          self.load(self.lazy_load)
          self.lazy_load = None

      samples = self.torch(samples)
      tgt = self.tgt_net(samples)
      prd = self.prd_net(samples)
      loss = F.mse_loss(tgt, prd)

      self.optimizer.zero_grad()
      loss.backward()
      self.optimizer.step()

  def save(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    if self.tgt_net is not None:
      torch.save({
        'tgt_state_dict': self.tgt_net.state_dict(),
        'prd_state_dict': self.prd_net.state_dict(),
        'opt_state_dict': self.optimizer.state_dict(),
      }, path)

  def load(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    if self.tgt_net is None and os.path.exists(path):
      self.lazy_load = save_folder
    else:
      checkpoint = torch.load(path)
      self.tgt_net.load_state_dict(checkpoint['tgt_state_dict'])
      self.prd_net.load_state_dict(checkpoint['prd_state_dict'])
      self.optimizer.load_state_dict(checkpoint['opt_state_dict'])


class FlowDensity(mrl.Module):
  """
  Flow Density model (in this case Real NVP). Similar structure to random density above
  """
  def __init__(self, item, optimize_every=2, batch_size=1000, lr=1e-3, num_layer_pairs=3, normalize=True):

    super().__init__('{}_flow'.format(item), required_agent_modules=['replay_buffer'], locals=locals())

    self.step = 0
    self.item = item
    self.num_layer_pairs = num_layer_pairs
    self.optimize_every = optimize_every
    self.batch_size = batch_size
    self.lazy_load = None
    self.flow_model = None
    self.dev = None
    self.lr = lr
    self.sample_mean = 0.
    self.sample_std = 1.
    self.normalize= normalize

  def _setup(self):
    assert isinstance(self.replay_buffer, OnlineHERBuffer)

  def _init_from_sample(self, x):
    input_size = x.shape[-1]
    self.input_channel = input_size
    if self.config.get('device'):
      self.dev = self.config.device
    elif self.dev is None:
      self.dev = 'cuda' if torch.cuda.is_available() else 'cpu'
    # Device=None is fine for default too based on network.py in realNVP
    self.flow_model = RealNVP(input_channel=self.input_channel, lr=self.lr, num_layer_pairs=self.num_layer_pairs, dev=self.dev)

  def evaluate_log_density(self, samples):
    assert self.ready, "ENSURE READY BEFORE EVALUATING LOG DENSITY"
    return self.flow_model.score_samples( (samples - self.sample_mean) / self.sample_std  )

  @property
  def ready(self):
    return self.flow_model is not None

  def _optimize(self, force=False):
    buffer = self.replay_buffer.buffer.BUFF['buffer_' + self.item]
    self.step +=1

    if force or (self.step % self.optimize_every == 0 and len(buffer)):
      sample_idxs = np.random.randint(len(buffer), size=self.batch_size)
      samples = buffer.get_batch(sample_idxs)
      if self.normalize:
        self.sample_mean = np.mean(samples, axis=0, keepdims=True)
        self.sample_std  = np.std(samples, axis=0, keepdims=True) + 1e-4
        samples = (samples - self.sample_mean) / self.sample_std

      # lazy load the model if not yet loaded
      if self.flow_model is None:
        self._init_from_sample(samples)
        if self.lazy_load is not None:
          self.load(self.lazy_load)
          self.lazy_load = None

      samples = self.torch(samples)
      #del self.flow_model
      #self.flow_model = RealNVP(input_channel=self.input_channel, lr=self.lr, num_layer_pairs=self.num_layer_pairs, dev=self.dev)
      self.flow_model.fit(samples, epochs=1)

  def save(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    if self.flow_model is not None:
      torch.save({
        'flow_model': self.flow_model,
      }, path)

  def load(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    if self.flow_model is None and os.path.exists(path):
      self.lazy_load = save_folder
    else:
      self.flow_model = torch.load(path)


"""
class DisagreementDensity(mrl.Module):
  #TODO
"""


import mrl
from mrl.utils.misc import soft_update, flatten_state
from mrl.modules.model import PytorchModel

import numpy as np
import torch
import torch.nn.functional as F
import os

class ActorPolicy(mrl.Module):
  """Used for DDPG / TD3 and other deterministic policy variants"""
  def __init__(self):
    super().__init__(
        'policy',
        required_agent_modules=[
            'actor', 'action_noise', 'env', 'replay_buffer'
        ],
        locals=locals())
  
  def _setup(self):
    self.use_actor_target = self.config.get('use_actor_target')

  def __call__(self, state, greedy=False):
    action_scale = self.env.max_action

    # initial exploration and intrinsic curiosity
    res = None
    if self.training:
      if self.config.get('initial_explore') and len(self.replay_buffer) < self.config.initial_explore:
        res = np.array([self.env.action_space.sample() for _ in range(self.env.num_envs)])
      elif hasattr(self, 'ag_curiosity'):
        state = self.ag_curiosity.relabel_state(state)
        
    state = flatten_state(state)  # flatten goal environments
    if hasattr(self, 'state_normalizer'):
      state = self.state_normalizer(state, update=self.training)

    if res is not None:
      return res

    state = self.torch(state)

    if self.use_actor_target:
      action = self.numpy(self.actor_target(state))
    else:
      action = self.numpy(self.actor(state))

    if self.training and not greedy:
      action = self.action_noise(action)
      if self.config.get('eexplore'):
        eexplore = self.config.eexplore
        if hasattr(self, 'ag_curiosity'):
          eexplore = self.ag_curiosity.go_explore * self.config.go_eexplore + eexplore
        mask = (np.random.random((action.shape[0], 1)) < eexplore).astype(np.float32)
        randoms = np.random.random(action.shape) * (2 * action_scale) - action_scale
        action = mask * randoms + (1 - mask) * action

    return np.clip(action, -action_scale, action_scale)


class StochasticActorPolicy(mrl.Module):
  """Used for SAC / learned action noise"""
  def __init__(self):
    super().__init__(
        'policy',
        required_agent_modules=[
            'actor', 'env', 'replay_buffer'
        ],
        locals=locals())

  def _setup(self):
    self.use_actor_target = self.config.get('use_actor_target')

  def __call__(self, state, greedy=False):
    action_scale = self.env.max_action

    # initial exploration and intrinsic curiosity
    res = None
    if self.training:
      if self.config.get('initial_explore') and len(self.replay_buffer) < self.config.initial_explore:
          res = np.array([self.env.action_space.sample() for _ in range(self.env.num_envs)])
      elif hasattr(self, 'ag_curiosity'):
        state = self.ag_curiosity.relabel_state(state)
        
    state = flatten_state(state)  # flatten goal environments
    if hasattr(self, 'state_normalizer'):
      state = self.state_normalizer(state, update=self.training)
    
    if res is not None:
      return res

    state = self.torch(state)

    if self.use_actor_target:
      action, _ = self.actor_target(state)
    else:
      action, _ = self.actor(state)
    action = self.numpy(action)

    if self.training and not greedy and self.config.get('eexplore'):
      eexplore = self.config.eexplore
      if hasattr(self, 'ag_curiosity'):
        eexplore = self.ag_curiosity.go_explore * self.config.go_eexplore + eexplore
      mask = (np.random.random((action.shape[0], 1)) < eexplore).astype(np.float32)
      randoms = np.random.random(action.shape) * (2 * action_scale) - action_scale
      action = mask * randoms + (1 - mask) * action
    
    return np.clip(action, -action_scale, action_scale)


class OffPolicyActorCritic(mrl.Module):
  """This is the standard DDPG"""

  def __init__(self):
    super().__init__(
        'algorithm',
        required_agent_modules=['actor','critic','replay_buffer', 'env'],
        locals=locals())

  def _setup(self):
    """Sets up actor/critic optimizers and creates target network modules"""

    self.targets_and_models = []

    # Actor setup
    actor_params = []
    self.actors = []
    for module in list(self.module_dict.values()):
      name = module.module_name
      if name.startswith('actor') and isinstance(module, PytorchModel):
        self.actors.append(module)
        actor_params += list(module.model.parameters())
        target = module.copy(name + '_target')
        target.model.load_state_dict(module.model.state_dict())
        # Freeze target networks with respect to optimizers (only update via polyak averaging)
        for p in target.model.parameters():
          p.requires_grad = False
        self.agent.set_module(name + '_target', target)
        self.targets_and_models.append((target.model, module.model))

    self.actor_opt = torch.optim.Adam(
        actor_params,
        lr=self.config.actor_lr,
        weight_decay=self.config.actor_weight_decay)
    
    self.actor_params = actor_params

    # Critic setup
    critic_params = []
    self.critics = []
    for module in list(self.module_dict.values()):
      name = module.module_name
      if name.startswith('critic') and isinstance(module, PytorchModel):
        self.critics.append(module)
        critic_params += list(module.model.parameters())
        target = module.copy(name + '_target')
        target.model.load_state_dict(module.model.state_dict())
        # Freeze target networks with respect to optimizers (only update via polyak averaging)
        for p in target.model.parameters():
          p.requires_grad = False
        self.agent.set_module(name + '_target', target)
        self.targets_and_models.append((target.model, module.model))

    self.critic_opt = torch.optim.Adam(
        critic_params,
        lr=self.config.critic_lr,
        weight_decay=self.config.critic_weight_decay)
    
    self.critic_params = critic_params

    self.action_scale = self.env.max_action

  def save(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    torch.save({
      'actor_opt_state_dict': self.actor_opt.state_dict(),
      'critic_opt_state_dict': self.critic_opt.state_dict()
    }, path)

  def load(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    checkpoint = torch.load(path)
    self.actor_opt.load_state_dict(checkpoint['actor_opt_state_dict'])
    self.critic_opt.load_state_dict(checkpoint['critic_opt_state_dict'])

  def _optimize(self):
    if len(self.replay_buffer) > self.config.warm_up:
      states, actions, rewards, next_states, gammas = self.replay_buffer.sample(
          self.config.batch_size)

      self.optimize_from_batch(states, actions, rewards, next_states, gammas)
      
      if self.config.opt_steps % self.config.target_network_update_freq == 0:
        for target_model, model in self.targets_and_models:
          soft_update(target_model, model, self.config.target_network_update_frac)

  def optimize_from_batch(self, states, actions, rewards, next_states, gammas):
    raise NotImplementedError('Subclass this!')


class DDPG(OffPolicyActorCritic):

  def optimize_from_batch(self, states, actions, rewards, next_states, gammas):

    with torch.no_grad():
      q_next = self.critic_target(next_states, self.actor_target(next_states))
      target = (rewards + gammas * q_next)
      target = torch.clamp(target, *self.config.clip_target_range)

    if hasattr(self, 'logger') and self.config.opt_steps % 1000 == 0:
      self.logger.add_histogram('Optimize/Target_q', target)
    
    q = self.critic(states, actions)
    critic_loss = F.mse_loss(q, target)

    self.critic_opt.zero_grad()
    critic_loss.backward()
      
    # Grad clipping
    if self.config.grad_norm_clipping > 0.:	
      for p in self.critic_params:
        clip_coef = self.config.grad_norm_clipping / (1e-6 + torch.norm(p.grad.detach()))
        if clip_coef < 1:
          p.grad.detach().mul_(clip_coef)
    if self.config.grad_value_clipping > 0.:
      torch.nn.utils.clip_grad_value_(self.critic_params, self.config.grad_value_clipping)

    self.critic_opt.step()

    for p in self.critic_params:
      p.requires_grad = False

    a = self.actor(states)
    if self.config.get('policy_opt_noise'):
      noise = torch.randn_like(a) * (self.config.policy_opt_noise * self.action_scale)
      a = (a + noise).clamp(-self.action_scale, self.action_scale)
      
    actor_loss = -self.critic(states, a)[:,-1].mean()
    if self.config.action_l2_regularization:
      actor_loss += self.config.action_l2_regularization * F.mse_loss(a / self.action_scale, torch.zeros_like(a))

    self.actor_opt.zero_grad()
    actor_loss.backward()
      
    # Grad clipping
    if self.config.grad_norm_clipping > 0.:	
      for p in self.actor_params:
        clip_coef = self.config.grad_norm_clipping / (1e-6 + torch.norm(p.grad.detach()))
        if clip_coef < 1:
          p.grad.detach().mul_(clip_coef)
    if self.config.grad_value_clipping > 0.:
      torch.nn.utils.clip_grad_value_(self.actor_params, self.config.grad_value_clipping)
      
    self.actor_opt.step()

    for p in self.critic_params:
      p.requires_grad = True


class TD3(OffPolicyActorCritic):

  def optimize_from_batch(self, states, actions, rewards, next_states, gammas):
    config = self.config

    with torch.no_grad():
      a_next_max = self.actor(next_states)
      noise = torch.randn_like(a_next_max) * (config.td3_noise * self.action_scale)
      noise = noise.clamp(-config.td3_noise_clip * self.action_scale,
                          config.td3_noise_clip * self.action_scale)
      a_next_max = (a_next_max + noise).clamp(-self.action_scale, self.action_scale)

      q1, q2 = self.critic_target(next_states, a_next_max), self.critic2_target(
          next_states, a_next_max)
      
      target = (rewards + gammas * torch.min(q1, q2))
      target = torch.clamp(target, *self.config.clip_target_range)

    if hasattr(self, 'logger') and self.config.opt_steps % 1000 == 0:
      self.logger.add_histogram('Optimize/Target_q', target)

    q1, q2 = self.critic(states, actions), self.critic2(states, actions)
    critic_loss = F.mse_loss(q1, target) + F.mse_loss(q2, target)

    self.critic_opt.zero_grad()
    critic_loss.backward()
      
    # Grad clipping
    if self.config.grad_norm_clipping > 0.:	
      torch.nn.utils.clip_grad_norm_(self.critic_params, self.config.grad_norm_clipping)
    if self.config.grad_value_clipping > 0.:
      torch.nn.utils.clip_grad_value_(self.critic_params, self.config.grad_value_clipping)

    self.critic_opt.step()

    if config.opt_steps % config.td3_delay == 0:
      for p in self.critic_params:
        p.requires_grad = False

      a = self.actor(states)
      if self.config.get('policy_opt_noise'):
        noise = torch.randn_like(a) * (config.policy_opt_noise * self.action_scale)
        a = (a + noise).clamp(-self.action_scale, self.action_scale)
      actor_loss = -torch.min(self.critic(states, a)[:,-1], self.critic2(states, a)[:,-1]).mean()
      if self.config.action_l2_regularization:
        actor_loss += self.config.action_l2_regularization * F.mse_loss(a / self.action_scale, torch.zeros_like(a))

      self.actor_opt.zero_grad()
      actor_loss.backward()
      
      # Grad clipping
      if self.config.grad_norm_clipping > 0.:	
        torch.nn.utils.clip_grad_norm_(self.actor_params, self.config.grad_norm_clipping)
      if self.config.grad_value_clipping > 0.:
        torch.nn.utils.clip_grad_value_(self.actor_params, self.config.grad_value_clipping)

      self.actor_opt.step()

      for p in self.critic_params:
        p.requires_grad = True


class SAC(OffPolicyActorCritic):

  def optimize_from_batch(self, states, actions, rewards, next_states, gammas):
    config = self.config

    with torch.no_grad():
      # Target actions come from *current* policy
      a_next, logp_next = self.actor(next_states)
      q1 = self.critic_target(next_states, a_next)
      q2 = self.critic2_target(next_states, a_next)
      target = rewards + gammas * (torch.min(q1, q2) - config.entropy_coef * logp_next)
      target = torch.clamp(target, *self.config.clip_target_range)

    if hasattr(self, 'logger') and self.config.opt_steps % 1000 == 0:
      self.logger.add_histogram('Optimize/Target_q', target)

    q1, q2 = self.critic(states, actions), self.critic2(states, actions)
    critic_loss = F.mse_loss(q1, target) + F.mse_loss(q2, target)

    self.critic_opt.zero_grad()
    critic_loss.backward()
      
    # Grad clipping
    if self.config.grad_norm_clipping > 0.:	
      torch.nn.utils.clip_grad_norm_(self.critic_params, self.config.grad_norm_clipping)
    if self.config.grad_value_clipping > 0.:
      torch.nn.utils.clip_grad_value_(self.critic_params, self.config.grad_value_clipping)

    self.critic_opt.step()

    for p in self.critic_params:
      p.requires_grad = False

    a, logp = self.actor(states)
    q = torch.min(self.critic(states, a), self.critic2(states, a))

    actor_loss = (config.entropy_coef * logp - q).mean()

    if self.config.action_l2_regularization:
      actor_loss += self.config.action_l2_regularization * F.mse_loss(a / self.action_scale, torch.zeros_like(a))

    self.actor_opt.zero_grad()
    actor_loss.backward()
      
    # Grad clipping
    if self.config.grad_norm_clipping > 0.:	
      torch.nn.utils.clip_grad_norm_(self.actor_params, self.config.grad_norm_clipping)
    if self.config.grad_value_clipping > 0.:
      torch.nn.utils.clip_grad_value_(self.actor_params, self.config.grad_value_clipping)

    self.actor_opt.step()

    for p in self.critic_params:
      p.requires_grad = True



"""
  In fusion-based ensemble, predictions from all base estimators are
  first aggregated as an average output. After then, the training loss is
  computed based on this average output and the ground-truth. The training
  loss is then back-propagated to all base estimators simultaneously.
"""


import torch
import torch.nn as nn
import torch.nn.functional as F

from ._base import BaseClassifier, BaseRegressor
from ._base import torchensemble_model_doc
from .utils import io
from .utils import set_module
from .utils import operator as op


__all__ = ["FusionClassifier", "FusionRegressor"]


@torchensemble_model_doc(
    """Implementation on the FusionClassifier.""", "model"
)
class FusionClassifier(BaseClassifier):
    def _forward(self, *x):
        """
        Implementation on the internal data forwarding in FusionClassifier.
        """
        # Average
        outputs = [estimator(*x) for estimator in self.estimators_]
        output = op.average(outputs)

        return output

    @torchensemble_model_doc(
        """Implementation on the data forwarding in FusionClassifier.""",
        "classifier_forward",
    )
    def forward(self, *x):
        output = self._forward(*x)
        proba = F.softmax(output, dim=1)

        return proba

    @torchensemble_model_doc(
        """Set the attributes on optimizer for FusionClassifier.""",
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name, **kwargs)

    @torchensemble_model_doc(
        """Set the attributes on scheduler for FusionClassifier.""",
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        super().set_scheduler(scheduler_name, **kwargs)

    @torchensemble_model_doc(
        """Set the training criterion for FusionClassifier.""",
        "set_criterion",
    )
    def set_criterion(self, criterion):
        super().set_criterion(criterion)

    @torchensemble_model_doc(
        """Implementation on the training stage of FusionClassifier.""", "fit"
    )
    def fit(
        self,
        train_loader,
        epochs=100,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):

        # Instantiate base estimators and set attributes
        for _ in range(self.n_estimators):
            self.estimators_.append(self._make_estimator())
        self._validate_parameters(epochs, log_interval)
        self.n_outputs = self._decide_n_outputs(train_loader)
        optimizer = set_module.set_optimizer(
            self, self.optimizer_name, **self.optimizer_args
        )

        # Set the scheduler if `set_scheduler` was called before
        if self.use_scheduler_:
            self.scheduler_ = set_module.set_scheduler(
                optimizer, self.scheduler_name, **self.scheduler_args
            )

        # Check the training criterion
        if not hasattr(self, "_criterion"):
            self._criterion = nn.CrossEntropyLoss()

        # Utils
        best_acc = 0.0
        total_iters = 0

        # Training loop
        for epoch in range(epochs):
            self.train()
            for batch_idx, elem in enumerate(train_loader):

                data, target = io.split_data_target(elem, self.device)
                batch_size = data[0].size(0)

                optimizer.zero_grad()
                output = self._forward(*data)
                loss = self._criterion(output, target)
                loss.backward()
                optimizer.step()

                # Print training status
                if batch_idx % log_interval == 0:
                    with torch.no_grad():
                        _, predicted = torch.max(output.data, 1)
                        correct = (predicted == target).sum().item()

                        msg = (
                            "Epoch: {:03d} | Batch: {:03d} | Loss:"
                            " {:.5f} | Correct: {:d}/{:d}"
                        )
                        self.logger.info(
                            msg.format(
                                epoch, batch_idx, loss, correct, batch_size
                            )
                        )
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "fusion/Train_Loss", loss, total_iters
                            )
                total_iters += 1

            # Validation
            if test_loader:
                self.eval()
                with torch.no_grad():
                    correct = 0
                    total = 0
                    for _, elem in enumerate(test_loader):
                        data, target = io.split_data_target(elem, self.device)
                        output = self.forward(*data)
                        _, predicted = torch.max(output.data, 1)
                        correct += (predicted == target).sum().item()
                        total += target.size(0)
                    acc = 100 * correct / total

                    if acc > best_acc:
                        best_acc = acc
                        if save_model:
                            io.save(self, save_dir, self.logger)

                    msg = (
                        "Epoch: {:03d} | Validation Acc: {:.3f}"
                        " % | Historical Best: {:.3f} %"
                    )
                    self.logger.info(msg.format(epoch, acc, best_acc))
                    if self.tb_logger:
                        self.tb_logger.add_scalar(
                            "fusion/Validation_Acc", acc, epoch
                        )

            # Update the scheduler
            if hasattr(self, "scheduler_"):
                self.scheduler_.step()

        if save_model and not test_loader:
            io.save(self, save_dir, self.logger)

    @torchensemble_model_doc(item="classifier_evaluate")
    def evaluate(self, test_loader, return_loss=False):
        return super().evaluate(test_loader, return_loss)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)


@torchensemble_model_doc("""Implementation on the FusionRegressor.""", "model")
class FusionRegressor(BaseRegressor):
    @torchensemble_model_doc(
        """Implementation on the data forwarding in FusionRegressor.""",
        "regressor_forward",
    )
    def forward(self, *x):
        # Average
        outputs = [estimator(*x) for estimator in self.estimators_]
        pred = op.average(outputs)

        return pred

    @torchensemble_model_doc(
        """Set the attributes on optimizer for FusionRegressor.""",
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name, **kwargs)

    @torchensemble_model_doc(
        """Set the attributes on scheduler for FusionRegressor.""",
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        super().set_scheduler(scheduler_name, **kwargs)

    @torchensemble_model_doc(
        """Set the training criterion for FusionRegressor.""",
        "set_criterion",
    )
    def set_criterion(self, criterion):
        super().set_criterion(criterion)

    @torchensemble_model_doc(
        """Implementation on the training stage of FusionRegressor.""", "fit"
    )
    def fit(
        self,
        train_loader,
        epochs=100,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):
        # Instantiate base estimators and set attributes
        for _ in range(self.n_estimators):
            self.estimators_.append(self._make_estimator())
        self._validate_parameters(epochs, log_interval)
        self.n_outputs = self._decide_n_outputs(train_loader)
        optimizer = set_module.set_optimizer(
            self, self.optimizer_name, **self.optimizer_args
        )

        # Set the scheduler if `set_scheduler` was called before
        if self.use_scheduler_:
            self.scheduler_ = set_module.set_scheduler(
                optimizer, self.scheduler_name, **self.scheduler_args
            )

        # Check the training criterion
        if not hasattr(self, "_criterion"):
            self._criterion = nn.MSELoss()

        # Utils
        best_loss = float("inf")
        total_iters = 0

        # Training loop
        for epoch in range(epochs):
            self.train()
            for batch_idx, elem in enumerate(train_loader):

                data, target = io.split_data_target(elem, self.device)

                optimizer.zero_grad()
                output = self.forward(*data)
                loss = self._criterion(output, target)
                loss.backward()
                optimizer.step()

                # Print training status
                if batch_idx % log_interval == 0:
                    with torch.no_grad():
                        msg = "Epoch: {:03d} | Batch: {:03d} | Loss: {:.5f}"
                        self.logger.info(msg.format(epoch, batch_idx, loss))
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "fusion/Train_Loss", loss, total_iters
                            )
                total_iters += 1

            # Validation
            if test_loader:
                self.eval()
                with torch.no_grad():
                    val_loss = 0.0
                    for _, elem in enumerate(test_loader):
                        data, target = io.split_data_target(elem, self.device)
                        output = self.forward(*data)
                        val_loss += self._criterion(output, target)
                    val_loss /= len(test_loader)

                    if val_loss < best_loss:
                        best_loss = val_loss
                        if save_model:
                            io.save(self, save_dir, self.logger)

                    msg = (
                        "Epoch: {:03d} | Validation Loss: {:.5f} |"
                        " Historical Best: {:.5f}"
                    )
                    self.logger.info(msg.format(epoch, val_loss, best_loss))
                    if self.tb_logger:
                        self.tb_logger.add_scalar(
                            "fusion/Validation_Loss", val_loss, epoch
                        )

            # Update the scheduler
            if hasattr(self, "scheduler_"):
                self.scheduler_.step()

        if save_model and not test_loader:
            io.save(self, save_dir, self.logger)

    @torchensemble_model_doc(item="regressor_evaluate")
    def evaluate(self, test_loader):
        return super().evaluate(test_loader)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)



import mrl
import gym
from mrl.replays.core.shared_buffer import SharedMemoryTrajectoryBuffer as Buffer
import numpy as np
import pickle
import os
from mrl.utils.misc import batch_block_diag

class OnlineHERBuffer(mrl.Module):

  def __init__(
      self,
      module_name='replay_buffer'
    ):
    """
    Buffer that does online hindsight relabeling.
    Replaces the old combo of ReplayBuffer + HERBuffer.
    """

    super().__init__(module_name, required_agent_modules=['env'], locals=locals())

    self.size = None
    self.goal_space = None
    self.buffer = None
    self.save_buffer = None
    
  def _setup(self):
    self.size = self.config.replay_size

    env = self.env
    if type(env.observation_space) == gym.spaces.Dict:
      observation_space = env.observation_space.spaces["observation"]
      self.goal_space = env.observation_space.spaces["desired_goal"]
    else:
      observation_space = env.observation_space

    items = [("state", observation_space.shape),
             ("action", env.action_space.shape), ("reward", (1,)),
             ("next_state", observation_space.shape), ("done", (1,))]

    if self.goal_space is not None:
      items += [("previous_ag", self.goal_space.shape), # for reward shaping
                ("ag", self.goal_space.shape), # achieved goal
                ("bg", self.goal_space.shape), # behavioral goal (i.e., intrinsic if curious agent)
                ("dg", self.goal_space.shape)] # desired goal (even if ignored behaviorally)

    self.buffer = Buffer(self.size, items)
    self._subbuffers = [[] for _ in range(self.env.num_envs)]
    self.n_envs = self.env.num_envs

    # HER mode can differ if demo or normal replay buffer
    if 'demo' in self.module_name:
      self.fut, self.act, self.ach, self.beh = parse_hindsight_mode(self.config.demo_her)
    else:
      self.fut, self.act, self.ach, self.beh = parse_hindsight_mode(self.config.her)

  def _process_experience(self, exp):
    if getattr(self, 'logger'):
      self.logger.add_tabular('Replay buffer size', len(self.buffer))
    done = np.expand_dims(exp.done, 1)  # format for replay buffer
    reward = np.expand_dims(exp.reward, 1)  # format for replay buffer
    action = exp.action

    if self.goal_space:
      state = exp.state['observation']
      next_state = exp.next_state['observation']
      previous_achieved = exp.state['achieved_goal']
      achieved = exp.next_state['achieved_goal']
      desired = exp.state['desired_goal']
      if hasattr(self, 'ag_curiosity') and self.ag_curiosity.current_goals is not None:
        behavioral = self.ag_curiosity.current_goals
        # recompute online reward
        reward = self.env.compute_reward(achieved, behavioral, {'s':state, 'ns':next_state}).reshape(-1, 1)
      else:
        behavioral = desired
      for i in range(self.n_envs):
        self._subbuffers[i].append([
            state[i], action[i], reward[i], next_state[i], done[i], previous_achieved[i], achieved[i],
            behavioral[i], desired[i]
        ])
    else:
      state = exp.state
      next_state = exp.next_state
      for i in range(self.n_envs):
        self._subbuffers[i].append(
            [state[i], action[i], reward[i], next_state[i], done[i]])

    for i in range(self.n_envs):
      if exp.trajectory_over[i]:
        trajectory = [np.stack(a) for a in zip(*self._subbuffers[i])]
        self.buffer.add_trajectory(*trajectory)
        self._subbuffers[i] = []

  def sample(self, batch_size, to_torch=True):
    if hasattr(self, 'prioritized_replay'):
      batch_idxs = self.prioritized_replay(batch_size)
    else:
      batch_idxs = np.random.randint(self.buffer.size, size=batch_size)

    if self.goal_space:
      if "demo" in self.module_name:
        has_config_her = self.config.get('demo_her')
      else:
        has_config_her = self.config.get('her')
      
      if has_config_her:

        if self.config.env_steps > self.config.future_warm_up:
          fut_batch_size, act_batch_size, ach_batch_size, beh_batch_size, real_batch_size = np.random.multinomial(
              batch_size, [self.fut, self.act, self.ach, self.beh, 1.])
        else:
          fut_batch_size, act_batch_size, ach_batch_size, beh_batch_size, real_batch_size = batch_size, 0, 0, 0, 0

        fut_idxs, act_idxs, ach_idxs, beh_idxs, real_idxs = np.array_split(batch_idxs, 
          np.cumsum([fut_batch_size, act_batch_size, ach_batch_size, beh_batch_size]))

        # Sample the real batch (i.e., goals = behavioral goals)
        states, actions, rewards, next_states, dones, previous_ags, ags, goals, _ =\
            self.buffer.sample(real_batch_size, batch_idxs=real_idxs)

        # Sample the future batch
        states_fut, actions_fut, _, next_states_fut, dones_fut, previous_ags_fut, ags_fut, _, _, goals_fut =\
          self.buffer.sample_future(fut_batch_size, batch_idxs=fut_idxs)

        # Sample the actual batch
        states_act, actions_act, _, next_states_act, dones_act, previous_ags_act, ags_act, _, _, goals_act =\
          self.buffer.sample_from_goal_buffer('dg', act_batch_size, batch_idxs=act_idxs)

        # Sample the achieved batch
        states_ach, actions_ach, _, next_states_ach, dones_ach, previous_ags_ach, ags_ach, _, _, goals_ach =\
          self.buffer.sample_from_goal_buffer('ag', ach_batch_size, batch_idxs=ach_idxs)

        # Sample the behavioral batch
        states_beh, actions_beh, _, next_states_beh, dones_beh, previous_ags_beh, ags_beh, _, _, goals_beh =\
          self.buffer.sample_from_goal_buffer('bg', beh_batch_size, batch_idxs=beh_idxs)

        # Concatenate the five
        states = np.concatenate([states, states_fut, states_act, states_ach, states_beh], 0)
        actions = np.concatenate([actions, actions_fut, actions_act, actions_ach, actions_beh], 0)
        ags = np.concatenate([ags, ags_fut, ags_act, ags_ach, ags_beh], 0)
        goals = np.concatenate([goals, goals_fut, goals_act, goals_ach, goals_beh], 0)
        next_states = np.concatenate([next_states, next_states_fut, next_states_act, next_states_ach, next_states_beh], 0)

        # Recompute reward online
        if hasattr(self, 'goal_reward'):
          rewards = self.goal_reward(ags, goals, {'s':states, 'ns':next_states}).reshape(-1, 1).astype(np.float32)
        else:
          rewards = self.env.compute_reward(ags, goals, {'s':states, 'ns':next_states}).reshape(-1, 1).astype(np.float32)

        if self.config.get('never_done'):
          dones = np.zeros_like(rewards, dtype=np.float32)
        elif self.config.get('first_visit_succ'):
          dones = np.round(rewards + 1.)
        else:
          raise ValueError("Never done or first visit succ must be set in goal environments to use HER.")
          dones = np.concatenate([dones, dones_fut, dones_act, dones_ach, dones_beh], 0)

        if self.config.sparse_reward_shaping:
          previous_ags = np.concatenate([previous_ags, previous_ags_fut, previous_ags_act, previous_ags_ach, previous_ags_beh], 0)
          previous_phi = -np.linalg.norm(previous_ags - goals, axis=1, keepdims=True)
          current_phi  = -np.linalg.norm(ags - goals, axis=1, keepdims=True)
          rewards_F = self.config.gamma * current_phi - previous_phi
          rewards += self.config.sparse_reward_shaping * rewards_F

      else:
        # Uses the original desired goals
        states, actions, rewards, next_states, dones, _ , _, _, goals =\
                                                    self.buffer.sample(batch_size, batch_idxs=batch_idxs)

      if self.config.slot_based_state:
        # TODO: For now, we flatten according to config.slot_state_dims
        I, J = self.config.slot_state_dims
        states = np.concatenate((states[:, I, J], goals), -1)
        next_states = np.concatenate((next_states[:, I, J], goals), -1)
      else:
        states = np.concatenate((states, goals), -1)
        next_states = np.concatenate((next_states, goals), -1)
      gammas = self.config.gamma * (1.-dones)

    elif self.config.get('n_step_returns') and self.config.n_step_returns > 1:
      states, actions, rewards, next_states, dones = self.buffer.sample_n_step_transitions(
        batch_size, self.config.n_step_returns, self.config.gamma, batch_idxs=batch_idxs
      )
      gammas = self.config.gamma**self.config.n_step_returns * (1.-dones)

    else:
      states, actions, rewards, next_states, dones = self.buffer.sample(
          batch_size, batch_idxs=batch_idxs)
      gammas = self.config.gamma * (1.-dones)

    if hasattr(self, 'state_normalizer'):
      states = self.state_normalizer(states, update=False).astype(np.float32)
      next_states = self.state_normalizer(
          next_states, update=False).astype(np.float32)
    
    if to_torch:
      return (self.torch(states), self.torch(actions),
            self.torch(rewards), self.torch(next_states),
            self.torch(gammas))
    else:
      return (states, actions, rewards, next_states, gammas)

  def __len__(self):
    return len(self.buffer)

  def save(self, save_folder):
    if self.config.save_replay_buf or self.save_buffer:
      state = self.buffer._get_state()
      with open(os.path.join(save_folder, "{}.pickle".format(self.module_name)), 'wb') as f:
        pickle.dump(state, f)

  def load(self, save_folder):
    load_path = os.path.join(save_folder, "{}.pickle".format(self.module_name))
    if os.path.exists(load_path):
      with open(load_path, 'rb') as f:
        state = pickle.load(f)
      self.buffer._set_state(state)
    else:
      self.logger.log_color('WARNING', 'Replay buffer is not being loaded / was not saved.', color='cyan')
      self.logger.log_color('WARNING', 'Replay buffer is not being loaded / was not saved.', color='red')
      self.logger.log_color('WARNING', 'Replay buffer is not being loaded / was not saved.', color='yellow')

def parse_hindsight_mode(hindsight_mode : str):
  if 'future_' in hindsight_mode:
    _, fut = hindsight_mode.split('_')
    fut = float(fut) / (1. + float(fut))
    act = 0.
    ach = 0.
    beh = 0.
  elif 'futureactual_' in hindsight_mode:
    _, fut, act = hindsight_mode.split('_')
    non_hindsight_frac = 1. / (1. + float(fut) + float(act))
    fut = float(fut) * non_hindsight_frac
    act = float(act) * non_hindsight_frac
    ach = 0.
    beh = 0.
  elif 'futureachieved_' in hindsight_mode:
    _, fut, ach = hindsight_mode.split('_')
    non_hindsight_frac = 1. / (1. + float(fut) + float(ach))
    fut = float(fut) * non_hindsight_frac
    act = 0.
    ach = float(ach) * non_hindsight_frac
    beh = 0.
  elif 'rfaa_' in hindsight_mode:
    _, real, fut, act, ach = hindsight_mode.split('_')
    denom = (float(real) + float(fut) + float(act) + float(ach))
    fut = float(fut) / denom
    act = float(act) / denom
    ach = float(ach) / denom
    beh = 0.
  elif 'rfaab_' in hindsight_mode:
    _, real, fut, act, ach, beh = hindsight_mode.split('_')
    denom = (float(real) + float(fut) + float(act) + float(ach) + float(beh))
    fut = float(fut) / denom
    act = float(act) / denom
    ach = float(ach) / denom
    beh = float(beh) / denom
  else:
    fut = 0.
    act = 0.
    ach = 0.
    beh = 0.

  return fut, act, ach, beh



import mrl
import numpy as np
import torch, torch.nn.functional as F
import os

class GoalEnvReward(mrl.Module):
  def __init__(self):
    """Wraps environment's compute reward function"""
    super().__init__(
        'goal_reward', required_agent_modules=['env'], locals=locals())

  def _setup(self):
    assert self.env.goal_env, "Environment must be a goal environment!"
    assert hasattr(self.env, 'compute_reward'), "Environment must have compute reward defined!"

  def __call__(self, achieved_goals, goals, info):
    return self.env.compute_reward(achieved_goals, goals, info)


class NeighborReward(mrl.Module):
  def __init__(self, max_neighbor_distance = 1, optimize_every = 5, batch_size = 1000, temperature = 1.):
    """Wraps environment's compute reward function. Should probably only be used for first-visit achievment."""
    super().__init__(
        'goal_reward', required_agent_modules=['replay_buffer', 'neighbor_embedding_network'], locals=locals())
    
    self.step = 0
    self.optimize_every = optimize_every
    self.batch_size = batch_size
    self.temperature = temperature

    if max_neighbor_distance != 1: # this is the number of steps from which to count two goals as neighbors.
      raise NotImplementedError

  def _setup(self):
    assert self.env.goal_env, "Environment must be a goal environment!"
    assert hasattr(self.env, 'compute_reward'), "Environment must have compute reward defined!"

    self.optimizer = torch.optim.Adam(
      self.neighbor_embedding_network.model.parameters(),
      lr=self.config.critic_lr, # just using critic hparams for now
      weight_decay=self.config.critic_weight_decay)

  def _optimize(self):
    pag_buffer = self.replay_buffer.buffer.BUFF.buffer_previous_ag
    ag_buffer = self.replay_buffer.buffer.BUFF.buffer_ag
    self.step +=1

    if self.step % self.optimize_every == 0 and len(ag_buffer):
      sample_idxs = np.random.randint(len(ag_buffer), size=self.batch_size)

      ags = ag_buffer.get_batch(sample_idxs)
      pos = pag_buffer.get_batch(sample_idxs)

      # mix it up to keep it symmetric for now...
      temp = ags[:len(ags) //2].copy()
      ags[:len(ags) //2] = pos[:len(ags) //2]
      pos[:len(ags) //2] = temp

      # get random negative samples by a 1 index roll
      neg = np.roll(pos, 1, axis=0)

      # move to torch
      ags = self.torch(ags)
      pos = self.torch(pos)
      neg = self.torch(neg)

      # get embeddings
      embs = self.neighbor_embedding_network(torch.cat((ags, pos, neg), dim=0))
      ags, pos, neg = torch.chunk(embs, 3)

      pos_logits = -self.temperature * torch.norm(ags - pos, dim = 1)
      neg_logits = -self.temperature * torch.norm(ags - neg, dim = 1)

      # use soft targets
      loss = F.binary_cross_entropy_with_logits(torch.exp(pos_logits), torch.ones_like(pos_logits) * 0.99) +\
             F.binary_cross_entropy_with_logits(torch.exp(neg_logits), torch.ones_like(pos_logits) * 0.01)


      self.logger.add_tabular('intrinsic_reward_loss', self.numpy(loss))

      # optimize
      self.optimizer.zero_grad()
      loss.backward()
      self.optimizer.step()

  def __call__(self, achieved_goals, goals, info):
    """Should return 0 for ags, gs that are predicted to be neighbors, -1 otherwise, as a numpy array"""
    ags = achieved_goals.reshape(-1, achieved_goals.shape[-1])
    dgs = goals.reshape(-1, achieved_goals.shape[-1])

    ags = self.torch(ags)
    dgs = self.torch(dgs)

    # get embeddings
    embs = self.neighbor_embedding_network(torch.cat((ags, dgs), dim=0))
    ags, dgs = torch.chunk(embs, 2)

    # predict whether ags and dgs are transition neighbors
    preds = torch.exp(-self.temperature * torch.norm(ags - dgs, dim = 1))

    return -self.numpy(preds < 0.5).astype(np.float32)

  def save(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    torch.save({
      'opt_state_dict': self.optimizer.state_dict()
    }, path)

  def load(self, save_folder : str):
    path = os.path.join(save_folder, self.module_name + '.pt')
    checkpoint = torch.load(path)
    self.optimizer.load_state_dict(checkpoint['opt_state_dict'])


  def load(self, save_folder):
    self._load_props(['random_process'], save_folder)


import abc
import copy
import torch
import logging
import warnings
import numpy as np
import torch.nn as nn

from . import _constants as const
from .utils.io import split_data_target
from .utils.logging import get_tb_logger


def torchensemble_model_doc(header="", item="model"):
    """
    A decorator on obtaining documentation for different methods in the
    ensemble. This decorator is modified from `sklearn.py` in XGBoost.

    Parameters
    ----------
    header: string
       Introduction to the decorated class or method.
    item : string
       Type of the docstring item.
    """

    def get_doc(item):
        """Return the selected item."""
        __doc = {
            "model": const.__model_doc,
            "seq_model": const.__seq_model_doc,
            "fit": const.__fit_doc,
            "predict": const.__predict_doc,
            "set_optimizer": const.__set_optimizer_doc,
            "set_scheduler": const.__set_scheduler_doc,
            "set_criterion": const.__set_criterion_doc,
            "classifier_forward": const.__classification_forward_doc,
            "classifier_evaluate": const.__classification_evaluate_doc,
            "regressor_forward": const.__regression_forward_doc,
            "regressor_evaluate": const.__regression_evaluate_doc,
        }
        return __doc[item]

    def adddoc(cls):
        doc = [header + "\n\n"]
        doc.extend(get_doc(item))
        cls.__doc__ = "".join(doc)
        return cls

    return adddoc


class BaseModule(nn.Module):
    """Base class for all ensembles.

    WARNING: This class cannot be used directly.
    Please use the derived classes instead.
    """

    def __init__(
        self,
        estimator,
        n_estimators,
        estimator_args=None,
        cuda=True,
        n_jobs=None,
    ):
        super(BaseModule, self).__init__()
        self.base_estimator_ = estimator
        self.n_estimators = n_estimators
        self.estimator_args = estimator_args

        if estimator_args and not isinstance(estimator, type):
            msg = (
                "The input `estimator_args` will have no effect since"
                " `estimator` is already an object after instantiation."
            )
            warnings.warn(msg, RuntimeWarning)

        self.device = torch.device("cuda" if cuda else "cpu")
        self.n_jobs = n_jobs
        self.logger = logging.getLogger()
        self.tb_logger = get_tb_logger()

        self.estimators_ = nn.ModuleList()
        self.use_scheduler_ = False

    def __len__(self):
        """
        Return the number of base estimators in the ensemble. The real number
        of base estimators may not match `self.n_estimators` because of the
        early stopping stage in several ensembles such as Gradient Boosting.
        """
        return len(self.estimators_)

    def __getitem__(self, index):
        """Return the `index`-th base estimator in the ensemble."""
        return self.estimators_[index]

    @abc.abstractmethod
    def _decide_n_outputs(self, train_loader):
        """Decide the number of outputs according to the `train_loader`."""

    def _make_estimator(self):
        """Make and configure a copy of `self.base_estimator_`."""

        # Call `deepcopy` to make a base estimator
        if not isinstance(self.base_estimator_, type):
            estimator = copy.deepcopy(self.base_estimator_)
        # Call `__init__` to make a base estimator
        else:
            # Without params
            if self.estimator_args is None:
                estimator = self.base_estimator_()
            # With params
            else:
                estimator = self.base_estimator_(**self.estimator_args)

        return estimator.to(self.device)

    def _validate_parameters(self, epochs, log_interval):
        """Validate hyper-parameters on training the ensemble."""

        if not epochs > 0:
            msg = (
                "The number of training epochs should be strictly positive"
                ", but got {} instead."
            )
            self.logger.error(msg.format(epochs))
            raise ValueError(msg.format(epochs))

        if not log_interval > 0:
            msg = (
                "The number of batches to wait before printing the"
                " training status should be strictly positive, but got {}"
                " instead."
            )
            self.logger.error(msg.format(log_interval))
            raise ValueError(msg.format(log_interval))

    def set_criterion(self, criterion):
        """Set the training criterion."""
        self._criterion = criterion

    def set_optimizer(self, optimizer_name, **kwargs):
        """Set the parameter optimizer."""
        self.optimizer_name = optimizer_name
        self.optimizer_args = kwargs

    def set_scheduler(self, scheduler_name, **kwargs):
        """Set the learning rate scheduler."""
        self.scheduler_name = scheduler_name
        self.scheduler_args = kwargs
        self.use_scheduler_ = True

    @abc.abstractmethod
    def forward(self, *x):
        """
        Implementation on the data forwarding in the ensemble. Notice
        that the input ``x`` should be a data batch instead of a standalone
        data loader that contains many data batches.
        """

    @abc.abstractmethod
    def fit(
        self,
        train_loader,
        epochs=100,
        log_interval=100,
        test_loader=None,
        save_model=True,
        save_dir=None,
    ):
        """
        Implementation on the training stage of the ensemble.
        """

    @torch.no_grad()
    def predict(self, *x):
        """Docstrings decorated by downstream ensembles."""
        self.eval()

        # Copy data
        x_device = []
        for data in x:
            if isinstance(data, torch.Tensor):
                x_device.append(data.to(self.device))
            elif isinstance(data, np.ndarray):
                x_device.append(torch.Tensor(data).to(self.device))
            else:
                msg = (
                    "The type of input X should be one of {{torch.Tensor,"
                    " np.ndarray}}."
                )
                raise ValueError(msg)

        pred = self.forward(*x_device)
        pred = pred.cpu()
        return pred


class BaseClassifier(BaseModule):
    """Base class for all ensemble classifiers.

    WARNING: This class cannot be used directly.
    Please use the derived classes instead.
    """

    def _decide_n_outputs(self, train_loader):
        """
        Decide the number of outputs according to the `train_loader`.
        The number of outputs equals the number of distinct classes for
        classifiers.
        """
        if hasattr(train_loader.dataset, "classes"):
            n_outputs = len(train_loader.dataset.classes)
        # Infer `n_outputs` from the dataloader
        else:
            labels = []
            for _, elem in enumerate(train_loader):
                _, target = split_data_target(elem, self.device)
                labels.append(target)
            labels = torch.unique(torch.cat(labels))
            n_outputs = labels.size(0)

        return n_outputs

    @torch.no_grad()
    def evaluate(self, test_loader, return_loss=False):
        """Docstrings decorated by downstream models."""
        self.eval()
        correct = 0
        total = 0
        loss = 0.0

        for _, elem in enumerate(test_loader):
            data, target = split_data_target(elem, self.device)
            output = self.forward(*data)
            _, predicted = torch.max(output.data, 1)
            correct += (predicted == target).sum().item()
            total += target.size(0)
            loss += self._criterion(output, target)

        acc = 100 * correct / total
        loss /= len(test_loader)

        if return_loss:
            return acc, float(loss)

        return acc


class BaseRegressor(BaseModule):
    """Base class for all ensemble regressors.

    WARNING: This class cannot be used directly.
    Please use the derived classes instead.
    """

    def _decide_n_outputs(self, train_loader):
        """
        Decide the number of outputs according to the `train_loader`.
        The number of outputs equals the number of target variables for
        regressors (e.g., `1` in univariate regression).
        """
        for _, elem in enumerate(train_loader):
            _, target = split_data_target(elem, self.device)
            if len(target.size()) == 1:
                n_outputs = 1  # univariate regression
            else:
                n_outputs = target.size(1)  # multivariate regression
            break

        return n_outputs

    @torch.no_grad()
    def evaluate(self, test_loader):
        """Docstrings decorated by downstream ensembles."""
        self.eval()
        loss = 0.0

        for _, elem in enumerate(test_loader):
            data, target = split_data_target(elem, self.device)
            output = self.forward(*data)
            loss += self._criterion(output, target)

        return float(loss) / len(test_loader)



import numpy as np
import pybullet as p
from matplotlib import pyplot as plt
from scipy import interpolate

from rrt_ml.environments.car_navigation_bullet_env import *
from rrt_ml.utilities.configs import *


# Set stats to true to track bullet stats
cfg = MasterConfig()
cfg.env.general.stats = True
cfg.env.general.gui = True
cfg.env.general.max_timestep = 480
cfg.env.car.f_val = 240

# Reset master to track stats
env = CarNavigationBulletEnv(cfg)
env.reset_master()

# Generate info
env.reset()
done = False
vel_rear_wheel = []
pos_front_wheel = []
while not done:
    _, vel, _, _ = p.getJointState(env.car.id, 3, physicsClientId=env.client)
    pos, _, _, _ = p.getJointState(env.car.id, 6, physicsClientId=env.client)
    vel_rear_wheel.append(vel)
    pos_front_wheel.append(pos)
    obs, reward, done, info = env.step([1, 1])

# Get variables
t = env.stats.bullet.time[0, :50, 0]
s = pos_front_wheel[:50]
v = vel_rear_wheel[:50]

# Interpolate
fs = interpolate.interp1d(t, s, kind='zero')
fv = interpolate.interp1d(t, v, kind='zero')
x = np.linspace(t[0], t[-1], 10000)
s_x = fs(x)
v_x = fv(x)

# Config plot
plt.rcParams['axes.grid'] = True

# Plot
fig, axs = plt.subplots(1, 2, squeeze=True)
axs[0].plot(x, v_x, label='Velocidade (rad/s)')
axs[0].set_xlabel('Tempo (s)')
axs[0].legend()
axs[1].plot(x, s_x, label='Posição (rad)', color='red')
axs[1].set_xlabel('Tempo (s)')
axs[1].legend()
fig.savefig('motor_graphs.png', dpi=300)

fig, ax = plt.subplots()
ax.plot(x, v_x, label='Velocidade (rad/s)')
ax.set_xlabel('Tempo (s)')
ax.legend()
fig.savefig("graficosMotor1.pdf", format="pdf", bbox_inches="tight")
fig.savefig("graficosMotor1.png", dpi=300, bbox_inches="tight")

fig, ax = plt.subplots()
ax.plot(x, s_x, label='Posição (rad)', color='red')
ax.set_xlabel('Tempo (s)')
ax.legend()
fig.savefig("graficosMotor2.pdf", format="pdf", bbox_inches="tight")
fig.savefig("graficosMotor2.png", dpi=300, bbox_inches="tight")
pass




import pickle
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.patches import Polygon
from matplotlib.collections import PatchCollection
from matplotlib.ticker import MaxNLocator

from rrt_ml.algorithms.sl import *
from rrt_ml.utilities.configs import *


def run_sl(cfg: MasterConfig):

    sample_learner = SL(cfg)

    # Train
    if cfg.general.is_train:
        sample_learner.train()

    # Test
    else:
        sample_learner.test()


def compare_hypers(cfg, name1, name2):
    """
    Compare two configs.
    :param name1: config 1
    :param name2: config 2
    """

    # Matplotlib params
    mpl.rcParams["font.size"] = 10

    # Experiments path
    base_path = Path(__file__).resolve().parents[2]
    exp_path = base_path / "experiments" / "sl"
    file_path = (
        base_path.parents[0] / "latex" / "figs" / "results" / "hiper_compare.png"
    )

    # Initialize figure
    fig, ax = plt.subplots(nrows=1, ncols=2, sharex="all", sharey="all")
    ax = ax.flatten()
    plt.tight_layout()

    # Axes counter
    ax_counter1 = 0
    ax_counter2 = 0
    ax_counter3 = 0

    path1 = exp_path / name1
    path2 = exp_path / name2
    with open(path1 / "stats.pickle", "rb") as file:
        stats1 = pickle.load(file)
    with open(path2 / "stats.pickle", "rb") as file:
        stats2 = pickle.load(file)

    # Plot
    for i, stats in enumerate([stats1, stats2]):

        # Aliases
        x1, y1, x2, y2 = stats["list_cond_obstacles_epoch"][49][0, :]
        ww = cfg.wall_width
        pw = cfg.passage_width

        # Left rectangle
        lower_left = Polygon(
            [
                (x1 - ww / 2, 0),
                (x1 + ww / 2, 0),
                (x1 + ww / 2, y1 - pw / 2),
                (x1 - ww / 2, y1 - pw / 2),
            ]
        )
        upper_left = Polygon(
            [
                (x1 - ww / 2, y1 + pw / 2),
                (x1 + ww / 2, y1 + pw / 2),
                (x1 + ww / 2, 10),
                (x1 - ww / 2, 10),
            ]
        )

        # Right rectangle
        lower_right = Polygon(
            [
                (x2 - ww / 2, 0),
                (x2 + ww / 2, 0),
                (x2 + ww / 2, y2 - pw / 2),
                (x2 - ww / 2, y2 - pw / 2),
            ]
        )
        upper_right = Polygon(
            [
                (x2 - ww / 2, y2 + pw / 2),
                (x2 + ww / 2, y2 + pw / 2),
                (x2 + ww / 2, 10),
                (x2 - ww / 2, 10),
            ]
        )
        patches = [lower_left, lower_right, upper_left, upper_right]

        ax[i].add_collection(
            PatchCollection(patches, edgecolors="k", facecolors="k")
        )  # type: plt.Axes

        # Add limits
        ax[ax_counter1].set_xlim(0, 10)
        ax[ax_counter1].set_ylim(0, 10)

        # Adjust aspect ratio
        ax[i].set_aspect("equal", adjustable="box")
        # ax[i].axis('equal')

        # Add initial and final states
        x1, y1, dx1, dy1, x2, y2, dx2, dy2 = stats["list_cond_states_epoch"][49][0, :]
        ax[ax_counter1].arrow(x1, y1, 0.3 * dx1, 0.3 * dy1, width=0.1, ec="g", fc="g")
        ax[ax_counter1].arrow(x2, y2, 0.3 * dx2, 0.3 * dy2, width=0.1, ec="r", fc="r")

        # Add generated states
        for state in stats["list_gen_states_epoch"][49][0:30]:
            norm = np.sqrt(state[2] ** 2 + state[3] ** 2)
            ax[ax_counter1].arrow(
                state[0],
                state[1],
                0.2 * state[2] / norm,
                0.2 * state[3] / norm,
                width=0.01,
                head_width=0.1,
                head_length=0.1,
                fc="b",
                ec="b",
            )

        # Set title
        name = name1 if i == 0 else name2
        ax[i].title.set_text(f"Hiperparâmetro {name}")

    # Save figure
    fig.savefig(file_path, dpi=600)


def hyperparameter_search_samples(cfg):
    """
    Graph for the various hyperparameters.
    :param cfg: experiment config.
    """

    # Matplotlib params
    mpl.rcParams["font.size"] = 14

    # Experiments path
    base_path = Path(__file__).resolve().parents[2]
    exp_path = base_path / "experiments" / "sl"
    file_path1 = base_path.parents[0] / "latex" / "figs" / "apdx" / "sl_hiper1.png"
    file_path2 = base_path.parents[0] / "latex" / "figs" / "apdx" / "sl_hiper2.png"
    file_path3 = base_path.parents[0] / "latex" / "figs" / "apdx" / "sl_hiper3.png"

    # Initialize figure 1
    fig1, axs1 = plt.subplots(nrows=4, ncols=2, sharex="all", sharey="all")
    fig1.set_size_inches(16, 22)
    axs1 = axs1.flatten()
    plt.tight_layout()

    # Initialize figure 2
    fig2, axs2 = plt.subplots(nrows=4, ncols=2, sharex="all", sharey="all")
    fig2.set_size_inches(16, 22)
    axs2 = axs2.flatten()
    plt.tight_layout()

    # Initialize figure 3
    fig3, axs3 = plt.subplots(nrows=4, ncols=2, sharex="all", sharey="all")
    fig3.set_size_inches(16, 22)
    axs3 = axs3.flatten()
    plt.tight_layout()

    # Axes counter
    ax_counter1 = 0
    ax_counter2 = 0
    ax_counter3 = 0

    # Go to experiment folder
    for param1 in [1, 2]:
        for param2 in [1, 2]:
            for param3 in [1, 2]:
                for param4 in [1, 2, 3]:
                    path = exp_path / (
                        "a" + str(param1) + str(param2) + str(param3) + str(param4)
                    )

                    # Load stats file
                    with open(path / "stats.pickle", "rb") as file:
                        stats = pickle.load(file)

                    # Aliases
                    x1, y1, x2, y2 = stats["list_cond_obstacles_epoch"][49][0, :]
                    ww = cfg.wall_width
                    pw = cfg.passage_width

                    # Left rectangle
                    lower_left = Polygon(
                        [
                            (x1 - ww / 2, 0),
                            (x1 + ww / 2, 0),
                            (x1 + ww / 2, y1 - pw / 2),
                            (x1 - ww / 2, y1 - pw / 2),
                        ]
                    )
                    upper_left = Polygon(
                        [
                            (x1 - ww / 2, y1 + pw / 2),
                            (x1 + ww / 2, y1 + pw / 2),
                            (x1 + ww / 2, 10),
                            (x1 - ww / 2, 10),
                        ]
                    )

                    # Right rectangle
                    lower_right = Polygon(
                        [
                            (x2 - ww / 2, 0),
                            (x2 + ww / 2, 0),
                            (x2 + ww / 2, y2 - pw / 2),
                            (x2 - ww / 2, y2 - pw / 2),
                        ]
                    )
                    upper_right = Polygon(
                        [
                            (x2 - ww / 2, y2 + pw / 2),
                            (x2 + ww / 2, y2 + pw / 2),
                            (x2 + ww / 2, 10),
                            (x2 - ww / 2, 10),
                        ]
                    )
                    patches = [lower_left, lower_right, upper_left, upper_right]

                    # Figure 1
                    if ax_counter1 < 8:

                        # Add patches
                        axs1[ax_counter1].add_collection(
                            PatchCollection(patches)
                        )  # type: plt.Axes

                        # Add limits
                        axs1[ax_counter1].set_xlim(0, 10)
                        axs1[ax_counter1].set_ylim(0, 10)

                        # Adjust aspect ratio
                        # axs1[ax_counter1].set_aspect('equal', adjustable='box')
                        # axs1[ax_counter1].axis('equal')

                        # Add initial and final states
                        x1, y1, dx1, dy1, x2, y2, dx2, dy2 = stats[
                            "list_cond_states_epoch"
                        ][49][0, :]
                        axs1[ax_counter1].arrow(
                            x1, y1, 0.3 * dx1, 0.3 * dy1, width=0.1, fc="g"
                        )
                        axs1[ax_counter1].arrow(
                            x2, y2, 0.3 * dx2, 0.3 * dy2, width=0.1, fc="r"
                        )

                        # Add generated states
                        for state in stats["list_gen_states_epoch"][49][0:30]:
                            norm = np.sqrt(state[2] ** 2 + state[3] ** 2)
                            axs1[ax_counter1].arrow(
                                state[0],
                                state[1],
                                0.2 * state[2] / norm,
                                0.2 * state[3] / norm,
                                width=0.01,
                                head_width=0.1,
                                head_length=0.1,
                                fc="k",
                            )

                        # Set title
                        axs1[ax_counter1].title.set_text(
                            f"Hiperparâmetro {param1}{param2}{param3}{param4}"
                        )

                        # Next subplot
                        ax_counter1 += 1

                    # Figure 2
                    else:

                        if ax_counter2 < 8:

                            # Add patches
                            axs2[ax_counter2].add_collection(
                                PatchCollection(patches)
                            )  # type: plt.Axes

                            # Add limits
                            axs2[ax_counter2].set_xlim(0, 10)
                            axs2[ax_counter2].set_ylim(0, 10)

                            # Adjust aspect ratio
                            # axs2[ax_counter2].set_aspect('equal')
                            # axs2[ax_counter2].axis('equal')

                            # Add initial and final states
                            x1, y1, dx1, dy1, x2, y2, dx2, dy2 = stats[
                                "list_cond_states_epoch"
                            ][49][0, :]
                            axs2[ax_counter2].arrow(
                                x1, y1, 0.3 * dx1, 0.3 * dy1, width=0.1, fc="g"
                            )
                            axs2[ax_counter2].arrow(
                                x2, y2, 0.3 * dx2, 0.3 * dy2, width=0.1, fc="r"
                            )

                            # Add generated states
                            for state in stats["list_gen_states_epoch"][49][0:30]:
                                norm = np.sqrt(state[2] ** 2 + state[3] ** 2)
                                axs2[ax_counter2].arrow(
                                    state[0],
                                    state[1],
                                    0.2 * state[2] / norm,
                                    0.2 * state[3] / norm,
                                    width=0.01,
                                    head_width=0.1,
                                    head_length=0.1,
                                    fc="k",
                                )

                            # Set title
                            axs2[ax_counter2].title.set_text(
                                f"Hiperparâmetro {param1}{param2}{param3}{param4}"
                            )

                            # Next subplot
                            ax_counter2 += 1

                        else:

                            # Add patches
                            axs3[ax_counter3].add_collection(
                                PatchCollection(patches)
                            )  # type: plt.Axes

                            # Add limits
                            axs3[ax_counter3].set_xlim(0, 10)
                            axs3[ax_counter3].set_ylim(0, 10)

                            # Adjust aspect ratio
                            # axs2[ax_counter2].set_aspect('equal')
                            # axs2[ax_counter2].axis('equal')

                            # Add initial and final states
                            x1, y1, dx1, dy1, x2, y2, dx2, dy2 = stats[
                                "list_cond_states_epoch"
                            ][49][0, :]
                            axs3[ax_counter3].arrow(
                                x1, y1, 0.3 * dx1, 0.3 * dy1, width=0.1, fc="g"
                            )
                            axs3[ax_counter3].arrow(
                                x2, y2, 0.3 * dx2, 0.3 * dy2, width=0.1, fc="r"
                            )

                            # Add generated states
                            for state in stats["list_gen_states_epoch"][49][0:30]:
                                norm = np.sqrt(state[2] ** 2 + state[3] ** 2)
                                axs3[ax_counter3].arrow(
                                    state[0],
                                    state[1],
                                    0.2 * state[2] / norm,
                                    0.2 * state[3] / norm,
                                    width=0.01,
                                    head_width=0.1,
                                    head_length=0.1,
                                    fc="k",
                                )

                            # Set title
                            axs3[ax_counter3].title.set_text(
                                f"Hiperparâmetro {param1}{param2}{param3}{param4}"
                            )

                            # Next subplot
                            ax_counter3 += 1

    # Save figure
    fig1.savefig(file_path1, dpi=600)
    fig2.savefig(file_path2, dpi=600)
    fig3.savefig(file_path3, dpi=600)


def hyperparameter_search_loss(cfg):
    """
    experiment config.
    :param cfg: experiment config.
    """

    # Matplotlib params
    mpl.rcParams["font.size"] = 14

    # Experiments path
    base_path = Path(__file__).resolve().parents[2]
    exp_path = base_path / "experiments" / "sl"
    file_path1 = base_path.parents[0] / "latex" / "figs" / "apdx" / "sl_hiper_loss1.png"
    file_path2 = base_path.parents[0] / "latex" / "figs" / "apdx" / "sl_hiper_loss2.png"
    file_path3 = base_path.parents[0] / "latex" / "figs" / "apdx" / "sl_hiper_loss3.png"

    # Initialize figure 1
    fig1, axs1 = plt.subplots(nrows=4, ncols=2, sharex="all")
    fig1.set_size_inches(16, 22)
    axs1 = axs1.flatten()
    # plt.tight_layout()

    # Initialize figure 2
    fig2, axs2 = plt.subplots(nrows=4, ncols=2, sharex="all")
    fig2.set_size_inches(16, 22)
    axs2 = axs2.flatten()
    # plt.tight_layout()

    # Initialize figure 3
    fig3, axs3 = plt.subplots(nrows=4, ncols=2, sharex="all")
    fig3.set_size_inches(16, 22)
    axs3 = axs3.flatten()
    # plt.tight_layout()

    # Axes counter
    ax_counter1 = 0
    ax_counter2 = 0
    ax_counter3 = 0

    # Go to experiment folder
    for param1 in [1, 2]:
        for param2 in [1, 2]:
            for param3 in [1, 2]:
                for param4 in [1, 2, 3]:
                    path = exp_path / (
                        "a" + str(param1) + str(param2) + str(param3) + str(param4)
                    )

                    # Load stats file
                    with open(path / "stats.pickle", "rb") as file:
                        stats = pickle.load(file)

                    # Figure 1
                    if ax_counter1 < 8:

                        # Plot loss with twin axes
                        axs1[ax_counter1].plot(
                            range(1, 51),
                            _smooth(stats["list_loss_recon_val_epoch"], 0.8),
                            color="b",
                        )
                        axs1[ax_counter1].tick_params(axis="y", labelcolor="b")
                        axs1[ax_counter1].yaxis.set_major_locator(
                            MaxNLocator(nbins=5, integer=True)
                        )
                        ax2 = axs1[ax_counter1].twinx()
                        ax2.plot(
                            range(1, 51),
                            _smooth(stats["list_loss_kl_val_epoch"], 0.8),
                            color="g",
                        )
                        ax2.tick_params(axis="y", labelcolor="g")
                        ax2.yaxis.set_major_locator(MaxNLocator(nbins=5, integer=True))
                        ax_counter1 += 1

                    # Figure 2
                    else:

                        if ax_counter2 < 8:

                            # Plot loss with twin axes
                            axs2[ax_counter2].plot(
                                range(1, 51),
                                _smooth(stats["list_loss_recon_val_epoch"], 0.8),
                                color="b",
                            )
                            axs2[ax_counter2].tick_params(axis="y", labelcolor="b")
                            axs2[ax_counter2].yaxis.set_major_locator(
                                MaxNLocator(nbins=5, integer=True)
                            )
                            ax2 = axs2[ax_counter2].twinx()
                            ax2.plot(
                                range(1, 51),
                                _smooth(stats["list_loss_kl_val_epoch"], 0.8),
                                color="g",
                            )
                            ax2.tick_params(axis="y", labelcolor="g")
                            ax2.yaxis.set_major_locator(
                                MaxNLocator(nbins=5, integer=True)
                            )
                            ax_counter2 += 1

                        else:

                            # Plot loss with twin axes
                            axs3[ax_counter3].plot(
                                range(1, 51),
                                _smooth(stats["list_loss_recon_val_epoch"], 0.8),
                                color="b",
                            )
                            axs3[ax_counter3].tick_params(axis="y", labelcolor="b")
                            axs3[ax_counter3].yaxis.set_major_locator(
                                MaxNLocator(nbins=5, integer=True)
                            )
                            ax2 = axs3[ax_counter3].twinx()
                            ax2.plot(
                                range(1, 51),
                                _smooth(stats["list_loss_kl_val_epoch"], 0.8),
                                color="g",
                            )
                            ax2.tick_params(axis="y", labelcolor="g")
                            ax2.yaxis.set_major_locator(
                                MaxNLocator(nbins=5, integer=True)
                            )
                            ax_counter3 += 1

    fig1.savefig(file_path1, dpi=600)
    fig2.savefig(file_path2, dpi=600)
    fig3.savefig(file_path3, dpi=600)


def _smooth(series, weight):
    last = series[0]  # First value in the plot (first timestep)
    smoothed = list()
    for point in series:
        smoothed_val = last * weight + (1 - weight) * point  # Calculate smoothed value
        smoothed.append(smoothed_val)  # Save it
        last = smoothed_val  # Anchor the last smoothed value
    return smoothed



import torch
import numpy as np
from mrl.utils.misc import AnnotatedAttrDict
from mrl.utils.schedule import LinearSchedule

default_dqn_config = lambda: AnnotatedAttrDict(
    device=('cuda' if torch.cuda.is_available() else 'cpu', 'torch device (cpu or gpu)'),
    gamma=(0.99, 'discount factor'),
    qvalue_lr=(1e-3, 'Q-value learning rate'),
    qvalue_weight_decay=(0., 'weight decay to apply to qvalue'),
    optimize_every=(2, 'how often optimize is called, in terms of environment steps'),
    batch_size=(1000, 'batch size for training the Q-values'),
    warm_up=(10000, 'minimum steps in replay buffer needed to optimize'),  
    initial_explore=(10000, 'whether to act randomly during warmup'), 
    grad_norm_clipping=(-1, 'gradient norm clipping (implemented as backward hook)'),
    grad_value_clipping=(-1, 'gradient value clipping'),
    random_action_prob=(LinearSchedule(1.0, 0.1, 1e4), 'Epsilon decay schedule'),
    target_network_update_frac=(0.005, 'polyak averaging coefficient for target networks'),
    target_network_update_freq=(2, 'how often to update target networks; NOTE: TD3 uses this too!'),
    clip_target_range=((-np.inf, np.inf), 'q/value targets are clipped to this range'),
    go_eexplore=(0.1, 'epsilon exploration bonus from each point of go explore, when using intrinsic curiosity'),
    go_reset_percent=(0.025, 'probability to reset epsiode early for each point of go explore, when using intrinsic curiosity'),
    overshoot_goal_percent=(False, 'if using instrinsic goals, should goal be overshot on success?'),
    direct_overshoots=(False, 'if using overshooting, should it be directed in a straight line?'),
    dg_score_multiplier=(1., 'if using instrinsic goals, score multiplier for goal candidates that are in DG distribution'),
    cutoff_success_threshold=(0.3, 0.7), # thresholds for decreasing/increasing the cutoff
    initial_cutoff=(-3, 'initial (and minimum) cutoff for intrinsic goal curiosity'),
    double_q=(False, 'Use Double DQN or not. Default: False'),
    activ=('gelu', 'activation to use for hidden layers in networks'),
    curiosity_beta=(-3., 'beta to use for curiosity_alpha module'),

    # Below are args to other modules (maybe should live in those modules?)
    seed=(0, 'random seed'),
    replay_size=(int(1e6), 'maximum size of replay buffer'),
    num_envs=(12, 'number of parallel envs to run'),
    log_every=(5000, 'how often to log things'),
    use_qvalue_target=(False, 'if true, use target network to act in the environment'),
    her=('futureactual_2_2', 'strategy to use for hindsight experience replay'),
    prioritized_mode=('none', 'buffer prioritization strategy'),
    future_warm_up=(20000, 'minimum steps in replay buffer needed to stop doing ONLY future sampling'),  
    sparse_reward_shaping=(0., 'coefficient of euclidean distance reward shaping in sparse goal envs'),
    n_step_returns=(1, 'if using n-step returns, how many steps?'),
    slot_based_state=(False, 'if state is organized by slot; i.e., [batch_size, num_slots, slot_feats]')
)

def dqn_config():
  config = default_dqn_config()
  config.gamma = 0.98
  config.qvalue_lr = 1e-4
  config.qvalue_weight_decay = 0.
  config.target_network_update_freq = 40
  config.target_network_update_frac = 0.05
  config.optimize_every = 2
  config.batch_size = 1000
  config.warm_up = 2500
  config.random_action_prob = LinearSchedule(1.0, 0.1, 1e4)
  config.replay_size = int(1e6)
  return config



import time
from multiprocessing import cpu_count

import matplotlib.ticker as ticker
import mrl
import pandas as pd
import pybullet as p
import seaborn as sns
from matplotlib import pyplot as plt
from mrl.algorithms.continuous_off_policy import ActorPolicy, DDPG
from mrl.configs.continuous_off_policy import protoge_config
from mrl.modules.action_noise import ContinuousActionNoise
from mrl.modules.curiosity import DensityAchievedGoalCuriosity
from mrl.modules.density import RawKernelDensity
from mrl.modules.env import EnvModule
from mrl.modules.eval import EpisodicEval
from mrl.modules.goal_reward import GoalEnvReward
from mrl.modules.logging import Logger
from mrl.modules.model import PytorchModel
from mrl.modules.normalizer import Normalizer, MeanStdNormalizer
from mrl.modules.train import StandardTrain
from mrl.replays.online_her_buffer import OnlineHERBuffer
from mrl.utils.misc import make_activ
from mrl.utils.networks import Actor, FCBody, Critic
from mrl.utils.random_process import GaussianProcess
from mrl.utils.schedule import ConstantSchedule
from rich.progress import track
from torch import nn

from rrt_ml.algorithms.base import *
from rrt_ml.environments.car_navigation_bullet_env import *
from rrt_ml.utilities.analytic import *
from rrt_ml.utilities.configs import *
from rrt_ml.utilities.formulas import *
from rrt_ml.utilities.hints import *
from rrt_ml.utilities.infos import *
from rrt_ml.utilities.paths import *
from rrt_ml.utilities.stats import *


class RL(Algorithm):

    def __init__(self, cfg: MasterConfig):
        """
        Initialize.
        :param cfg: configuration
        """

        super(RL, self).__init__(cfg)
        self.agent = None  # type: None | mrl.Agent
        self.eval_env = None  # type: None | CarNavigationBulletEnv
        self.mrl_config = None

        # Setup
        self._setup()

    def train(self):
        """
        Train agent.
        """

        # No need if already trained
        if self.training_over:
            return

        # Get total num of epochs
        epochs_total = int(self.cfg.rl.train.n_timesteps // self.cfg.rl.val.interval)

        # Train loop
        for epoch_n in track(range(self.epoch + 1, epochs_total), "Training RL agent..."):
            # Train for a number of timesteps
            self.agent.train(num_steps=self.cfg.rl.val.interval)

            # Evaluate with MRL (random resets) and custom (deterministic reset)
            rand_val_reward = np.mean(self.agent.eval(num_episodes=self.cfg.rl.val.n_episodes).rewards)
            env_stats, det_val_reward = self._get_eval_stats()
            train_timestep_num = epoch_n * self.cfg.rl.val.interval

            # Get and save epoch info
            epoch_info = RLEpochInfo.new(det_val_reward, rand_val_reward, env_stats, train_timestep_num)
            self._set_attrs_after_epoch(epoch_info)

            # Save
            self.save_stats(epoch_info)
            self.save_checkpoint(epoch_info)

            # Log
            self.log_console(epoch_info)
            self.log_tensorboard(epoch_info)

        # Set training flag and save attrs
        self.training_over = True
        self._set_save_attributes()

    def test(self):
        """
        Test RL.
        """

        self._test_plot_entropy()
        self._test_plot_rewards()
        self._test_plot_progress()
        self._test_plot_horizon_success_rate()
        self._test_plot_multiple_problems()
        self._test_plot_compare_mpcrs()
        self._test_plot_compare_rs()
        self._test_plot_compare_rs_mpcrs()
        self._test_chasing()

    def save_checkpoint(self, epoch_info: 'RLEpochInfo'):
        """
        Save checkpoint. Equal to base method except for last line.
        """

        # Save attributes
        self._set_save_attributes()

        # Save agent if improved deterministic and random evaluations
        if self.model_curr_loss > self.model_best_loss:
            # Log
            self.console.print(f"\n[red bold underline]Total reward increased from "
                               f"{self.model_best_loss:.2f} to {self.model_curr_loss:.2f}... Saving model...[/]")

            # Save agent
            self.agent.save("checkpoint")

    def save_stats(self, epoch_info: 'RLEpochInfo'):
        """
        Set up stats to save.
        """

        self.stats.env_stats.append(epoch_info.env_stats)
        self.stats.det_val_rewards.append(epoch_info.det_val_reward)
        self.stats.rand_val_rewards.append(epoch_info.rand_val_reward)
        self.stats.total_val_rewards.append(epoch_info.det_val_reward + epoch_info.rand_val_reward)
        self.stats.train_timestep_nums.append(epoch_info.train_timestep_num)

        self.stats.save_to_file(self.path_stats)

    def log_console(self, epoch_info: 'RLEpochInfo'):
        """
        Log to console.
        :param epoch_info: current epoch info
        """

        tt = epoch_info.train_timestep_num
        rr = epoch_info.rand_val_reward
        dr = epoch_info.det_val_reward
        self.console.print(
            f"\n[blue bold underline]Train Timestep:[/blue bold underline] [blue]{tt}[/blue]\t"
            f"[cyan bold underline]Current Rewards (rand/det):[/cyan bold underline] [cyan]{rr}/{dr}[/cyan]\t"
        )

    def log_tensorboard(self, epoch_info: 'RLEpochInfo'):
        """
        For consistency only, handled by MRL.
        """
        pass

    def get_distance_data(self, num_episodes=100000):
        """
        Get distance data for training supervised models.
        """

        # Create files and folders
        train_file_path, val_file_path, test_file_path = self._get_distance_data_files_paths()

        # Set agent to eval mode
        self.agent.eval_mode()

        # Create env with modified config and reset master to get all stats
        cfg = self._get_distance_env_cfg()
        env = CarNavigationBulletEnv(cfg)

        # Loop episodes
        ep_counter = 0
        for i in range(num_episodes):

            # Get data
            data = []
            l_obs = []
            self._get_episode_info_recursive(env, l_obs, data)

            # Increment counter
            ep_counter += len(data)

            # Save to file
            if ep_counter % 2 == 0:

                # Save batch
                self._set_save_distance_batch(data, train_file_path, val_file_path, test_file_path)

                # Log
                self.console.print(
                    f"\n[blue bold underline]Generating distance data...[/blue bold underline]\t"
                    f"[cyan bold underline]Number of episodes:[/cyan bold underline] [cyan]{ep_counter}[/cyan]\t"
                )

    def get_action(self, obs: dict[str, Vector]) -> Vector2:
        """
        Get agent prediction.
        :return: action
        """

        self.agent.eval_mode()

        return self.agent.policy(obs).flatten()

    def _get_distance_data_files_paths(self):
        """
        Create files and folders to hold data.
        :returns: train and validation csv paths
        """

        # Set paths
        folder_path = Paths().data_rl_distance / self.cfg.general.config_name_or_prefix
        train_file_path = folder_path / 'train.csv'
        val_file_path = folder_path / 'val.csv'
        test_file_path = folder_path / 'test.csv'

        # Create files and folders
        if not folder_path.exists():
            folder_path.mkdir()
        if not train_file_path.exists():
            with open(str(train_file_path), "w") as _:
                pass
            with open(str(val_file_path), "w") as _:
                pass
            with open(str(test_file_path), "w") as _:
                pass

        return train_file_path, val_file_path, test_file_path

    def _get_distance_env_cfg(self):
        """
        Change env config to generate distance data.
        :return: new master config
        """

        # Change config
        cfg = MasterConfig.load_from_experiment('rl', self.cfg.rrt.names.rl)
        cfg.env.car.pose = 0
        cfg.env.target.pose = 5
        cfg.env.general.max_timestep = 100
        cfg.env.general.seed = None
        cfg.env.general.stats = True
        cfg.env.general.gui = False

        return cfg

    def _get_eval_stats(self) -> tuple[EnvStats, float]:
        """
        Eval agent on episodes with deterministic resetting and capture stats.
        :return: stats and accumulated reward
        """

        # Set agent to eval mode (don't use dropout, etc)
        self.agent.eval_mode()

        # Master reset to set the seed and reset deterministically
        self.eval_env.reset_master()

        # Episodes loops
        sum_rewards = 0
        for i in range(self.cfg.rl.val.n_episodes):
            done = False
            obs = self.eval_env.reset()
            while not done:
                obs, reward, done, _ = self.eval_env.step(self.agent.policy(obs).flatten())
                sum_rewards += reward

        # Get stats
        return self.eval_env.stats, sum_rewards / self.cfg.rl.val.n_episodes

    def _get_episode_info_recursive(self, env: CarNavigationBulletEnv, l_obs: list[dict, ...], data: list[np.ndarray, ...]):
        """
        Get episode info recursively.
        :param env:
        :param l_obs:
        :param data:
        :return:
        """

        # If there are no list of observations we need to get it
        if len(l_obs) == 0:

            # If there is data we need to return it
            if len(data) > 0:
                return data

            # Place target at random pose
            env.reset_master()
            env.reset()

            # Place car at origin
            env.car.set_pose([0, 0, 0])

            # Get target obs
            obs = env._get_observation()
            car_state = obs['achieved_goal']
            target_state = obs['desired_goal']

            # Play episode
            done, done_info = env._get_done_indicator(env.compute_reward(car_state, target_state))
            info = {'done_info': done_info}
            while not done:

                # Policy action
                action = self.agent.policy(obs).flatten()

                # Step
                obs, reward, done, info = env.step(action)
                
                # Add to list of obs
                l_obs.append(obs)

            # Add to database
            success = info['done_info']['success']
            data.append(np.concatenate((
                target_state,
                np.array([success]),
                np.array([env.stats.get_time_to_reach(0)]),
                np.array([env.timestep]),
                np.array([env.stats.get_distance_traveled(0)]))
            ))

            self._get_episode_info_recursive(env, l_obs, data)

        else:

            # Reset to get stats and zero joints
            env.reset_master()
            env.reset()

            # Transform to origin
            env.car.set_pose([0, 0, 0])
            env.target.set_pose(transform_to_origin(l_obs.pop(0)))
            
            # Get new target pose
            obs = env._get_observation()
            car_state = obs['achieved_goal']
            target_state = obs['desired_goal']

            # Play episode
            done, done_info = env._get_done_indicator(env.compute_reward(car_state, target_state))
            info = {'done_info': done_info}
            while not done:

                # Policy action
                action = self.agent.policy(obs).flatten()

                # Step
                obs, reward, done, info = env.step(action)

            # Add to database
            success = info['done_info']['success']
            data.append(np.concatenate((
                    target_state,
                    np.array([success]),
                    np.array([env.stats.get_time_to_reach(0)]),
                    np.array([env.timestep]),
                    np.array([env.stats.get_distance_traveled(0)]))
                ))

            # Replay until there are no intermediate observations
            self._get_episode_info_recursive(env, l_obs, data)
            
            return data

    def _set_attrs_after_epoch(self, epoch_info):
        """
        Set basic attributes after an epoch o training.
        :param epoch_info: current epoch info
        """

        # Update iteration
        self.epoch = int(100 * epoch_info.train_timestep_num / self.cfg.rl.train.n_timesteps)

        # Update current loss
        self.model_curr_loss = epoch_info.det_val_reward + epoch_info.rand_val_reward

        # Check best loss so far
        if len(self.stats.total_val_rewards) > 1:
            self.model_best_loss = max(self.stats.total_val_rewards)

    def _set_save_distance_batch(self, data: list, train_file_path: str, val_file_path: str, test_file_path: str):
        """
        Save batch of distance data
        :param data: list of rows with x and y pairs
        :param train_file_path: path to train_file
        :param val_file_path: path to val_file
        """

        # Ignore warnings
        self._set_ignore_static_warnings()

        # Generate 10% of data as validation and 20% as test
        rand = np.random.rand()
        if rand < 0.1:
            with open(str(val_file_path), "a") as file:
                # noinspection PyTypeChecker
                np.savetxt(file, np.array(data), delimiter=",")
        elif rand < 0.25:
            with open(str(test_file_path), "a") as file:
                # noinspection PyTypeChecker
                np.savetxt(file, np.array(data), delimiter=",")
        else:
            with open(str(train_file_path), "a") as file:
                # noinspection PyTypeChecker
                np.savetxt(file, np.array(data), delimiter=",")

    def _test_chasing(self):
        """
        Test chasing different poses.
        """

        # Env
        cfg = self.cfg.copy(deep=True)
        cfg.env.general.gui = True
        cfg.env.car.pose = [0, 0, 0]
        cfg.env.target.pose = 1
        cfg.env.reward.epsilon = -0.15
        cfg.env.general.max_timestep = 10000
        env = CarNavigationBulletEnv(cfg)
        obs = env.reset()

        # Loop all episodes
        for _ in range(100):

            # Loop
            time.sleep(0.5)
            done = False
            pose = np.random.uniform([-1, -1, 0], [1, 1, 2*np.pi])
            env.target.set_pose(pose)
            while not done:
                obs, reward, done, info = env.step(self.get_action(obs))
                time.sleep(1/45)

    def _test_plot_compare_mpcrs(self):
        """
        Compare RL and MPC with Reed-Shepp paths.
        """

        # Return if figure exists
        path = self.path_figs / 'rl_comparacao_rsmpc.png'
        if path.exists():
            return

        # Change config
        cfg = self.cfg.copy(deep=True)
        cfg.env.general.seed = 1
        cfg.env.general.gui = True
        cfg.env.general.stats = True
        cfg.env.target.pose = 1.5
        cfg.env.car.pose = [0, 0, 0]

        # Initialize env and reset master
        env = CarNavigationBulletEnv(cfg)
        env.reset_master()

        # Initialize figure
        fig, axs = plt.subplots(nrows=2, ncols=3, sharex='col', sharey='row', figsize=(12, 9.6))
        fig.subplots_adjust(hspace=-0.2)

        # Episode loop for RL
        for ep in range(3):

            # Get episode info for RL
            done, obs = False, env.reset()
            while not done:
                obs, reward, done, info = env.step(self.get_action(obs))

            # Get car info to place ghosts and actions
            car_poses = env.stats.bullet.car_poses[ep, ...]
            car_steers = env.stats.bullet.car_steers[ep, ...]

            # Remove all zeros rows
            car_poses = car_poses[~np.all(car_poses == 0, axis=1)]
            car_steers = car_steers[~np.all(car_steers == 0, axis=1)]

            # Get indexes to place ghosts
            interval = cfg.env.test.general.ghost_interval
            n_timesteps = car_poses.shape[0]
            idxs = np.arange(0, n_timesteps, interval)

            # Place ghosts
            for pose, steer in zip(car_poses[idxs, :], car_steers[idxs, :]):
                env._set_place_ghost_car(pose, steer)

            # Place car at final pose
            env.car.set_pose(car_poses[-1, :])

            # Take screenshot and clear ghosts
            img = env._get_image_current_top_view()
            env._set_remove_ghosts()

            # Plot on axis
            axs[0, ep].imshow(img, origin='lower', extent=cfg.env.test.none.img_extent)

            # Get distance traveled and time
            length = env.stats.get_distance_traveled(ep)
            time = env.stats.get_time_to_reach(ep)
            axs[0, ep].set_title(f'Agente: {time:.2f}s / {length:.2f}m')

            # Set y axis
            if ep == 0:
                axs[0, ep].set_ylabel('y [m]')

        # Now get results for RSMPC, change config and re-init env
        cfg.env.car.f_val = 240
        cfg.env.general.max_timestep = 2000
        p.disconnect(physicsClientId=env.client)
        del env
        env = CarNavigationBulletEnv(cfg)
        env.reset_master()

        # Change config and initialize MPCRS
        rsmpc = MPCRS(cfg)

        # Episode loop for MPCRS
        for ep in range(3):

            # Get episode info for RL
            done, obs = False, env.reset()
            while not done:
                action = rsmpc.get_action(obs)
                obs, reward, done, info = env.step(action)

            # Get car info to place ghosts and actions
            car_poses = env.stats.bullet.car_poses[ep, ...]
            car_steers = env.stats.bullet.car_steers[ep, ...]

            # Remove all zeros rows
            car_poses = car_poses[~np.all(car_poses == 0, axis=1)]
            car_steers = car_steers[~np.all(car_steers == 0, axis=1)]

            # Get indexes to place ghosts
            interval = cfg.env.test.general.ghost_interval
            n_timesteps = car_poses.shape[0]
            idxs = np.arange(0, n_timesteps, interval)

            # Place ghosts
            for pose, steer in zip(car_poses[idxs, :], car_steers[idxs, :]):
                env._set_place_ghost_car(pose, steer)

            # Place car at final pose
            env.car.set_pose(car_poses[-1, :])

            # Take screenshot and clear ghosts
            img = env._get_image_current_top_view()
            env._set_remove_ghosts()

            # Plot on axis
            axs[1, ep].imshow(img, origin='lower', extent=cfg.env.test.none.img_extent)

            # Get distance traveled and time
            length = env.stats.get_distance_traveled(ep)
            time = env.stats.get_time_to_reach(ep)
            axs[1, ep].set_title(f'CPM: {time:.2f}s / {length:.2f}m')

            # Set x axis
            axs[1, ep].set_xlabel('x [m]')

            # Set y axis
            if ep == 0:
                axs[1, ep].set_ylabel('y [m]')

        # Delete env
        p.disconnect(physicsClientId=env.client)

        # Save figure
        fig.savefig(str(path), dpi=600, bbox_inches='tight')

    def _test_plot_compare_rs(self):
        """
        Compare RL path with RS path.
        """

        # Return if figure exists
        path = self.path_figs / 'rl_comparacao_rs.png'
        if path.exists():
            return

        # Change config
        cfg = self.cfg.copy(deep=True)
        cfg.env.general.seed = 1
        cfg.env.general.gui = True
        cfg.env.general.stats = True
        cfg.env.target.pose = 1.5
        cfg.env.car.pose = [0, 0, 0]
        cfg.rrt.rs.curvature = 2.5

        # Initialize env and reset master
        env = CarNavigationBulletEnv(cfg)
        env.reset_master()

        # Initialize figure
        fig, axs = plt.subplots(nrows=2, ncols=3, sharex='col', sharey='row', figsize=(12, 9.6))
        fig.subplots_adjust(hspace=-0.2)

        # Episode loop for RL
        ep = 0
        for i, _ in enumerate(axs):
            for j, _ in enumerate(axs[i, :]):

                # Reset env
                done, obs = False, env.reset()

                # Get RS path length before changing env
                rs = MPCRS(cfg)
                rs._set_rs_path(obs)
                length_rs = rs.get_distance(Node(env.car.get_pose()), Node(env.target.get_pose()))

                while not done:
                    obs, reward, done, info = env.step(self.get_action(obs))

                # Get car info to place ghosts and actions
                car_poses = env.stats.bullet.car_poses[ep, ...]
                car_steers = env.stats.bullet.car_steers[ep, ...]

                # Remove all zeros rows
                car_poses = car_poses[~np.all(car_poses == 0, axis=1)]
                car_steers = car_steers[~np.all(car_steers == 0, axis=1)]

                # Get indexes to place ghosts
                interval = cfg.env.test.general.ghost_interval
                n_timesteps = car_poses.shape[0]
                idxs = np.arange(0, n_timesteps, interval)

                # Place ghosts
                for pose, steer in zip(car_poses[idxs, :], car_steers[idxs, :]):
                    env._set_place_ghost_car(pose, steer)

                # Place car at final pose
                env.car.set_pose(car_poses[-1, :])

                # Take screenshot and clear ghosts
                img = env._get_image_current_top_view()
                env._set_remove_ghosts()

                # Plot on axis
                axs[i, j].imshow(img, origin='lower', extent=cfg.env.test.none.img_extent)

                # Plot RS path
                axs[i, j].plot(rs.rs_xs, rs.rs_ys, color='yellow', lw=2)

                # Get distance traveled
                length_rl = env.stats.get_distance_traveled(ep)

                # Set title as lengths
                axs[i, j].set_title(f'Agente: {length_rl:.2f}m\nRS: {length_rs:.2f}m')

                # Set x axis
                if i == 1:
                    axs[i, j].set_xlabel('x [m]')

                # Set y axis
                if j in [0, 3]:
                    axs[i, j].set_ylabel('y [m]')

                # Increment episode counter
                ep += 1

        # Delete env
        p.disconnect(physicsClientId=env.client)

        # Save figure
        fig.savefig(path, dpi=600, bbox_inches='tight')

    def _test_plot_compare_rs_mpcrs(self):
        """
        Compare RL, RS and MPC path lengths.
        """

        # Return if figure exists
        path = self.path_figs / 'rl_comparacao_rs_rsmpc.png'
        if path.exists():
            return

        # Constants
        n_trials = 20

        # Change config
        cfg = self.cfg.copy(deep=True)
        cfg.env.general.seed = 1
        cfg.env.general.gui = True
        cfg.env.general.stats = True
        cfg.env.target.pose = 1.5

        # Initialize dataframe
        df = pd.DataFrame({
            'Critério': [],
            'Custo': [],
            'Semente': [],
            'Abordagem': []
        })

        # Prepare for MPC
        cfg.env.car.f_val = 240
        cfg.env.general.max_timestep = 2000

        # Episode loop for MPC
        env = CarNavigationBulletEnv(cfg)
        env.reset_master()
        mpcrs = MPCRS(cfg)
        success_ep_nums = []
        i = -1
        while len(success_ep_nums) < n_trials:

            # Increment
            i += 1

            # Get episode info for RL
            done, obs, info = False, env.reset(), {}
            while not done:
                action = mpcrs.get_action(obs)
                obs, reward, done, info = env.step(action)

            if info['done_info']['success']:
                success_ep_nums.append(i)

            # Add to dataframe
            df.loc[len(df)] = ['Distância', env.stats.get_distance_traveled(i), i, 'RS+CPM']
            df.loc[len(df)] = ['Tempo', env.stats.get_time_to_reach(i), i, 'RS+CPM']

        # Episode loop for RL
        p.disconnect(physicsClientId=env.client)
        del env
        cfg = self.cfg.copy(deep=True)
        cfg.env.general.seed = 1
        cfg.env.general.gui = True
        cfg.env.general.stats = True
        cfg.env.target.pose = 1.5
        env = CarNavigationBulletEnv(cfg)
        env.reset_master()
        i = -1
        for j in range(success_ep_nums[-1]+1):

            # Reset before maybe continuing
            done, obs = False, env.reset()

            # Increment
            i += 1

            # Continue if not success for MPC
            if i not in success_ep_nums:
                continue

            # Get episode info for RL
            while not done:
                obs, reward, done, info = env.step(self.get_action(obs))

            # Add to dataframe
            if i in success_ep_nums:
                df.loc[len(df)] = ['Distância', env.stats.get_distance_traveled(i), i, 'Agente']
                df.loc[len(df)] = ['Tempo', env.stats.get_time_to_reach(i), i, 'Agente']

        # Prepare for RS
        cfg.rrt.rs.curvature = 2.5
        cfg.rrt.rs.step_size = 1/240
        p.disconnect(physicsClientId=env.client)
        del env
        env = CarNavigationBulletEnv(cfg)

        # Change config and initialize RS
        env.reset_master()
        mpcrs = MPCRS(cfg)

        # Episode loop for RS
        i = -1
        for j in range(success_ep_nums[-1] + 1):

            # Reset before maybe continuing
            done, obs = False, env.reset()

            # Increment
            i += 1

            # Continue if not success for MPC
            if i not in success_ep_nums:
                continue

            # Get episode info for RL
            mpcrs._set_rs_path(obs)
            node_from = Node(env.car.get_pose())
            node_to = Node(env.target.get_pose())

            if i in success_ep_nums:
                df.loc[len(df)] = ['Distância', mpcrs.get_distance(node_from, node_to), i, 'RS']
                df.loc[len(df)] = ['Tempo', (1/240)*len(mpcrs.rs_xs), i, 'RS']

        # Delete env
        p.disconnect(physicsClientId=env.client)

        # Rename members
        df2 = df.replace('Distância', 'Distância (m)')
        df2 = df2.replace('Tempo', 'Tempo (s)')

        # Save figure
        sns.set(font_scale=1.2)
        sns.set_style('whitegrid')
        ax = sns.barplot(data=df2, x="Critério", y="Custo", hue='Abordagem')
        l = ax.legend()
        l.set_title('')
        ax.get_figure().savefig(str(path), dpi=600, bbox_inches='tight')

    def _test_plot_entropy(self):
        """
        Get entropy plot.
        """

        # Return if figure exists
        path = self.path_figs / 'rl_entropia.png'
        if path.exists():
            return

        # Get entropy csv path
        csv_path = next(self.path_figs.parent.glob('*entropy*'))

        # Read as dataframe
        df = pd.read_csv(str(csv_path))

        # Plot
        fig, ax = plt.subplots()
        ax.plot(df['step'][7:], df['Explore/ag_kde_entropy'][7:])
        ax.grid(True)
        ax.set_xlabel('Passos de tempo')
        ax.set_ylabel('Entropia')
        fig.savefig(str(path), dpi=600, bbox_inches='tight')

    def _test_plot_horizon_success_rate(self):
        """
        Plot success rate for different horizons.
        """

        # Return if figure exists
        path = self.path_figs / 'rl_horizonte_taxa_sucesso.png'
        if path.exists():
            return

        # Env without gui to go faster and car pose is always [0, 0, 0]
        cfg = self.cfg.copy(deep=True)
        cfg.env.general.gui = False
        cfg.env.car.pose = [0, 0, 0]

        # Initialize distances
        distances = [0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0,
                     5.5, 6.0, 6.5, 7.0, 7.5, 8.0, 8.5, 9.0, 9.5, 10.0]

        # Loop through distances
        successes = []
        for d in distances:

            # Change env config distance and initialize
            cfg.env.target.pose = d
            env = CarNavigationBulletEnv(cfg)

            # Loop all episodes
            success = 0
            for _ in range(100):

                # Episode loop
                done, obs, info = False, env.reset(), {}
                while not done:
                    obs, reward, done, info = env.step(self.get_action(obs))

                # Check done reason
                if info['done_info']['success']:
                    success += 1

            # Delete env
            del env

            # Add to list
            successes.append(success)

            # Log
            self.console.print(f"\n[red bold underline]Done testing distance {d}...[/]")

        # Initialize figure and plot
        fig, ax = plt.subplots()  # type: plt.Figure, plt.Axes
        ax.bar([str(d) for d in distances], successes, color='blue', edgecolor='blue')

        # Rotate x-axis labels to fit better
        for label in ax.get_xticklabels():
            label.set_rotation(45)

        # Change y-axis to percent formatting
        ax.yaxis.set_major_formatter(ticker.PercentFormatter())

        # Add grid, axis labels
        ax.grid(True)
        ax.set_xlabel('c [m]')
        ax.set_ylabel('Taxa de sucesso')

        # Save
        fig.savefig(str(path), dpi=600, bbox_inches='tight')

    def _test_plot_rewards(self):
        """
        Plot test reward curves.
        """

        # Return if figure exists
        path = self.path_figs / 'rl_recompensa.png'
        if path.exists():
            return

        # Get constants
        timesteps = self.stats.train_timestep_nums
        rewards = self.stats.total_val_rewards

        # Plot
        fig, ax = plt.subplots()
        ax.plot(timesteps, rewards)
        ax.grid(True)
        ax.set_xlabel('Passos de tempo')
        ax.set_ylabel('Recompensa')
        fig.savefig(path, dpi=600, bbox_inches='tight')

    def _test_plot_multiple_problems(self):
        """
        Plot agent actions in an episode.
        """

        # Return if figure exists
        path = self.path_figs / 'rl_diversos_problemas.png'
        if path.exists():
            return

        # Change config
        cfg = self.cfg.copy(deep=True)
        cfg.env.general.seed = 0
        cfg.env.general.gui = True
        cfg.env.general.stats = True
        cfg.env.target.pose = 1.5
        cfg.env.car.pose = [0, 0, 0]

        # Initialize env and reset master
        env = CarNavigationBulletEnv(cfg)
        env.reset_master()

        # Initialize figure
        fig, axs = plt.subplots(nrows=2, ncols=3, sharex='col', sharey='row', figsize=(12, 9.6))
        fig.subplots_adjust(hspace=-0.2)
        axs = axs.flatten()

        # Episode loop
        for ep in range(6):

            # Get episode info
            done, obs = False, env.reset()
            while not done:
                obs, reward, done, info = env.step(self.get_action(obs))

            # Get car info to place ghosts and actions
            car_poses = env.stats.bullet.car_poses[ep, ...]
            car_steers = env.stats.bullet.car_steers[ep, ...]

            # Remove all zeros rows
            car_poses = car_poses[~np.all(car_poses == 0, axis=1)]
            car_steers = car_steers[~np.all(car_steers == 0, axis=1)]

            # Get indexes to place ghosts
            interval = cfg.env.test.general.ghost_interval
            n_timesteps = car_poses.shape[0]
            idxs = np.arange(0, n_timesteps, interval)

            # Place ghosts
            for pose, steer in zip(car_poses[idxs, :], car_steers[idxs, :]):
                env._set_place_ghost_car(pose, steer)

            # Place car at final pose
            env.car.set_pose(car_poses[-1, :])

            # Take screenshot and clear ghosts
            img = env._get_image_current_top_view()
            env._set_remove_ghosts()

            # Plot on axis
            axs[ep].imshow(img, origin='lower', extent=cfg.env.test.none.img_extent)

            # Set titles
            axs[ep].set_title(f'Cenário {ep + 1}')

            # Set axis labels
            if ep in [0, 3]:
                axs[ep].set_ylabel('y [m]')
            if ep in [3, 4, 5]:
                axs[ep].set_xlabel('x [m]')

        # Disconnect env
        p.disconnect(physicsClientId=env.client)

        # Save figure
        fig.savefig(str(path), dpi=600, bbox_inches='tight')

    def _test_plot_progress(self):
        """
        Plot policy trajectories over training time.
        """

        # Return if figure exists
        path = self.path_figs / 'rl_progresso.png'
        if path.exists():
            return

        # Constants
        ep_idx = 0
        train_timesteps_idx = [1, 2, 3, 4, 5, 90]

        # Change config before creating env
        cfg = self.cfg.copy(deep=True)
        cfg.env.general.gui = True

        # Initialize env to get screenshot
        env = CarNavigationBulletEnv(cfg)

        # Initialize figure
        fig, axs = plt.subplots(2, 3, sharex='col', sharey='row', figsize=(12, 9.6))
        fig.subplots_adjust(hspace=-0.2)
        axs = axs.flatten()

        # Loop to get subplots
        for i, idx in enumerate(train_timesteps_idx):

            # Get target pose
            desired_goal = self.stats.env_stats[idx].mdp.desired_goals[ep_idx, 0, ...]
            target_pose = state4_to_pose3(desired_goal)

            # Place target at desired pose
            env.target.set_pose(target_pose)

            # Get car info to place ghosts
            car_poses = self.stats.env_stats[idx].bullet.car_poses[ep_idx, ...]
            car_steers = self.stats.env_stats[idx].bullet.car_steers[ep_idx, ...]

            # Remove all zeros rows
            car_poses = car_poses[~np.all(car_poses == 0, axis=1)]
            car_steers = car_steers[~np.all(car_steers == 0, axis=1)]

            # Get indexes to place ghosts
            interval = cfg.env.test.general.ghost_interval
            n_timesteps = car_poses.shape[0]
            idxs = np.arange(0, n_timesteps, interval)

            # Place ghosts
            for pose, steer in zip(car_poses[idxs, :], car_steers[idxs, :]):
                env._set_place_ghost_car(pose, steer)

            # Place car at final pose
            env.car.set_pose(car_poses[-1, :])

            # Take screenshot
            img = env._get_image_current_top_view()

            # Plot on axis
            axs[i].imshow(img, origin='lower', extent=cfg.env.test.none.img_extent)

            # Set title
            axs[i].set_title(f'Época {idx}')

            # Set axis labels
            if i in [0, 3]:
                axs[i].set_ylabel('y [m]')
            if i in [3, 4, 5]:
                axs[i].set_xlabel('x [m]')

            # Clear ghosts
            env._set_remove_ghosts()

        # Disconnect env
        p.disconnect(physicsClientId=env.client)

        # Save figure
        fig.savefig(str(path), dpi=600, bbox_inches='tight')

    def _setup(self):
        """
        Setup.
        """

        # Set up agent must be called before everything
        self._setup_agent()

        # Set up base settings
        self._setup_paths()
        self._setup_folders()
        self._setup_save_config_to_file()
        self._setup_attrs_to_save()
        self._setup_init_base_attrs()
        self._setup_init_model()
        self._setup_checkpoint()
        self._setup_stats()
        self._setup_console()
        self._setup_tensorboard()

        # Set up RL settings
        self._setup_eval_env()
        self._setup_just_use_policy()

    def _setup_agent(self):
        """
        Setup mega default and merge with user configuration.
        """

        # Load default mega config
        config = protoge_config()

        # Differences from protoge
        config.action_l2_regularization = self.cfg.rl.actor.l2
        config.action_noise = self.cfg.rl.exploration.noise
        config.activ = self.cfg.rl.net.activ
        config.batch_size = self.cfg.rl.train.batch_size
        config.eexplore = self.cfg.rl.exploration.epsilon
        config.grad_value_clipping = self.cfg.rl.net.grad_value_clipping
        config.initial_explore = self.cfg.rl.exploration.initial
        config.layers = self.cfg.rl.net.layers
        config.replay_size = self.cfg.rl.train.replay_size
        config.target_network_update_freq = self.cfg.rl.target.update_freq
        config.warm_up = self.cfg.rl.exploration.warm_up
        config.her = self.cfg.rl.general.her
        config.optimize_every = self.cfg.rl.train.optimize_every
        config.replay_size = self.cfg.rl.train.replay_size
        config.target_network_update_freq = self.cfg.rl.target.update_freq

        # Experiments folder
        config.parent_folder = Paths().experiments_rl

        # Agent name
        config.agent_name = str(self.cfg.general.config_name_or_prefix)

        # Parallel
        if self.cfg.rl.train.n_envs is None:
            config.num_envs = max(cpu_count() - 2, 1)
        else:
            config.num_envs = self.cfg.rl.train.n_envs
        if self.cfg.rl.val.n_envs is None:
            config.num_eval_envs = 1
        else:
            config.num_eval_envs = self.cfg.rl.val.n_envs

        # No parallel if testing?
        if not self.cfg.general.is_train:
            config.num_envs = 1
            config.num_eval_envs = 1

        # Train
        config.train_timestep = self.cfg.rl.train.n_timesteps

        # Setup and add basic modules to the config
        config.update(
            dict(
                trainer=StandardTrain(),
                evaluation=EpisodicEval(),
                policy=ActorPolicy(),
                logger=Logger(),
                state_normalizer=Normalizer(MeanStdNormalizer()),
                replay=OnlineHERBuffer(),
            )
        )

        # Discount factor
        if config.gamma < 1.0:
            config.clip_target_range = (np.round(-(1 / (1 - config.gamma)), 2), 0.0)
        if config.gamma == 1:
            config.clip_target_range = (np.round(-self.cfg.env.general.max_timestep - 5, 2), 0.0)

        # Prioritized experience replay
        config.prioritized_mode = "none"

        # Curiosity - ag density estimation
        config.ag_kde_tophat = RawKernelDensity(
            "ag",
            optimize_every=100,
            samples=10000,
            kernel="tophat",
            bandwidth=0.2,
            tag="_tophat",
        )
        config.ag_kde = RawKernelDensity(
            "ag",
            optimize_every=1,
            samples=10000,
            kernel="gaussian",
            bandwidth=0.1,
            log_entropy=True,
        )
        config.ag_curiosity = DensityAchievedGoalCuriosity(
            max_steps=self.cfg.env.general.max_timestep,
            num_sampled_ags=100,
            use_qcutoff=True,
            keep_dg_percent=-0.1,
        )

        # Actor noise?
        config.action_noise = ContinuousActionNoise(
            GaussianProcess, std=ConstantSchedule(config.action_noise)
        )

        # Off-policy model
        config.algorithm = DDPG()

        # Change target pose and gui (for train and val is always off)
        cfg = self.cfg.deep_copy_change('env.target.pose', self.cfg.rl.val.target_pose)
        cfg = cfg.deep_copy_change('env.general.gui', False)

        # Set up train environment
        def train_env_fn():
            return CarNavigationBulletEnv(cfg)

        # Set up train environment
        def val_env_fn():
            return CarNavigationBulletEnv(cfg)

        config.train_env = EnvModule(
            train_env_fn, num_envs=config.num_envs
        )
        config.eval_env = EnvModule(
            val_env_fn,
            num_envs=config.num_eval_envs,
            name="eval_env",
        )

        # Setup env success & done
        config.first_visit_succ = True
        config.first_visit_done = False

        # Setup algorithm
        config.algorithm = DDPG()

        # Setup layer normalization
        layer_norm_or_not = nn.LayerNorm if self.cfg.rl.net.layer_norm else nn.Identity

        # Setup and add the networks to the config
        e = config.eval_env
        config.actor = PytorchModel(
            "actor",
            lambda: Actor(
                FCBody(
                    e.state_dim + e.goal_dim,
                    self.cfg.rl.net.layers,
                    layer_norm_or_not,
                    make_activ(config.activ),
                ),
                e.action_dim,
                e.max_action,
            ),
        )
        config.critic = PytorchModel(
            "critic",
            lambda: Critic(
                FCBody(
                    e.state_dim + e.goal_dim + e.action_dim,
                    self.cfg.rl.net.layers,
                    layer_norm_or_not,
                    make_activ(config.activ),
                ),
                1,
            ),
        )

        # Intrinsic reward
        config.goal_reward = GoalEnvReward()

        # Return agent and complete config
        self.agent = mrl.config_to_agent(config)
        self.mrl_config = config

    def _setup_best_rewards(self):
        """
        Set up best rewards.
        """

        self.best_det_val_reward = -np.inf
        self.best_rand_val_reward = -np.inf

    def _setup_checkpoint(self):
        """
        Set up checkpoint.
        """

        # Check if there is a checkpoint
        self.load_checkpoint = self._get_checkpoint_exists_indicator()

        if self.load_checkpoint:
            self._set_load_attributes()

    def _setup_eval_env(self):
        """
        Set up evaluation env with deterministic resetting.
        """

        # Change config before creating env
        cfg = self.cfg.deep_copy_change('env.general.gui', self.cfg.rl.val.gui)
        cfg = cfg.deep_copy_change('env.general.stats', self.cfg.rl.val.stats)

        self.eval_env = CarNavigationBulletEnv(cfg=cfg)

    def _setup_init_base_attrs(self):
        """
        Initialize base attributes.
        """

        self.epoch = 0
        self.model_best_loss = -np.inf
        self.model_curr_loss = -np.inf

    def _setup_just_use_policy(self):
        """
        If loading algorithm just to use the policy we delete all envs.
        """

        # Disconnect all envs used for training, validation and testing
        if self.cfg.rl.general.just_use_policy:

            # Disconnect from eval env
            p.disconnect(self.eval_env.client)

            # Disconnect from train env module
            for env in self.agent.env.env.envs:
                p.disconnect(env.env.client)

            # Disconnect from eval env module
            for env in self.agent.eval_env.env.envs:
                p.disconnect(env.env.client)

    def _setup_stats(self):
        """
        Set up stats.
        """

        if self.load_checkpoint:
            self.stats = RLStats.load_from_file(self.path_stats)
        else:
            self.stats = RLStats.new(self.cfg)
            self.stats.save_to_file(self.path_stats)

    def _setup_tensorboard(self):
        """
        For consistency only, handled by MRL.
        """

        pass



import torch
import pytest
import numpy as np
import torch.nn as nn
from torch.utils.data import TensorDataset, DataLoader

import torchensemble
from torchensemble.utils.logging import set_logger


set_logger("pytest_adversarial_training")


X_train = torch.Tensor(np.array(([1, 1], [2, 2], [3, 3], [4, 4])))
y_train_clf = torch.LongTensor(np.array(([0, 0, 1, 1])))


# Base estimator
class MLP_clf(nn.Module):
    def __init__(self):
        super(MLP_clf, self).__init__()
        self.linear1 = nn.Linear(2, 2)
        self.linear2 = nn.Linear(2, 2)

    def forward(self, X):
        X = X.view(X.size()[0], -1)
        output = self.linear1(X)
        output = self.linear2(output)
        return output


def test_adversarial_training_range():
    """
    This unit test checks the input range check in adversarial_training.py.
    """
    model = torchensemble.AdversarialTrainingClassifier(
        estimator=MLP_clf, n_estimators=2, cuda=False
    )

    model.set_optimizer("Adam")

    # Prepare data
    train = TensorDataset(X_train, y_train_clf)
    train_loader = DataLoader(train, batch_size=2)

    # Training
    with pytest.raises(ValueError) as excinfo:
        model.fit(train_loader)
    assert "input range of samples passed to adversarial" in str(excinfo.value)
